import os
import glob
import yaml
import qtc.utils.datetime_utils as dtu
from qtc.ext.logging import set_logger
logger = set_logger()


def extract_datetime_from_path(path, dt_index=None, filename_sep='.', timezone=None):
    """
    >>> import qtc.utils.file_system_utils as fsu
    >>> fsu.extract_datetime_from_path(path='$HOME/earnings.20211004.csv')
    datetime.datetime(2021, 10, 4, 0, 0)
    """
    dt = None
    try:
        components = os.path.basename(path).split(filename_sep)
        if dt_index is None:
            for component in components:
                try:
                    dt = dtu.parse_datetime(component, ret_as_timestamp=False, timezone=timezone)
                    break
                except:
                    continue
        else:
            dt = dtu.parse_datetime(components[dt_index], ret_as_timestamp=False, timezone=timezone)
    except Exception as e:
        logger.error(f'{path} not recognized as containing datetime component with '
                     f'dt_index={dt_index}, filename_sep={filename_sep}')
        return None

    return dt


def is_earlier(path, target_dt=None, dt_index=None, filename_sep='.', timezone=None):
    """
    >>> import qtc.utils.file_system_utils as fsu
    >>> fsu.is_earlier(path='${HOME}/market_data.20200901-210228.csv',
                       target_dt='20200901-210245')
    True
    >>> fsu.is_earlier(path='${HOME}/market_data_.20200901-210228.csv',
                       target_dt='20200901-210100')
    False
    """

    target_dt = dtu.normalize_dt(dt=target_dt, timezone=timezone)
    if target_dt is None:
        raise Exception(f'target_dt={target_dt} cannot be recognized as a valid datetime !')

    dt = extract_datetime_from_path(path=path, dt_index=dt_index, filename_sep=filename_sep, timezone=timezone)
    if dt is None:
        return False

    return dt <= target_dt


def list_datetime_subfolders(folder,
                             ret_abs_path=True, ascending=True, dt_no_later_than=None,
                             timezone=None):
    """
    >>> from qtc.ext.path import sysequity_dir
    >>> folder = sysequity_dir('Reports', 'Archive')
    >>> import qtc.utils.file_system_utils as fsu
    >>> fsu.list_datetime_subfolders(folder=folder)[:5]
    >>> fsu.list_datetime_subfolders(folder=folder, ret_abs_path=False)[:5]
    ['20160104', '20160105', '20160106', '20160107', '20160108']
    >>> fsu.list_datetime_subfolders(folder=folder, dt_no_later_than=20160105)
    >>> fsu.list_datetime_subfolders(folder=folder, ret_abs_path=False, ascending=False)[-5:]
    ['20160108', '20160107', '20160106', '20160105', '20160104']
    """
    def is_valid_datetime_folder(path, dt_no_later_than=None):
        if not os.path.isdir(path):
            return False

        dt = None
        try:
            subfolder = os.path.basename(path)
            dt = dtu.parse_datetime(subfolder, ret_as_timestamp=False)
        except Exception as e:
            logger.debug(f"Skipping '{subfolder}' since it is not a date subfolder! ")
            return False

        if dt_no_later_than is None:
            return True

        dt_no_later_than = dtu.normalize_dt(dt=dt_no_later_than, timezone=timezone)
        if dt_no_later_than is None:
            logger.warn(f'dt_no_later_than={dt_no_later_than} cannot be recognized as a valid datetime !')
            return True

        return dt <= dt_no_later_than

    if folder is None or folder.strip()=='' or not os.path.exists(folder):
        return None

    if dt_no_later_than is not None:
        dt_no_later_than = dtu.normalize_dt(dt=dt_no_later_than, timezone=timezone)

    ret = [os.path.join(folder, subfolder) if ret_abs_path else subfolder
           for subfolder in os.listdir(folder) if is_valid_datetime_folder(os.path.join(folder, subfolder),
                                                                           dt_no_later_than=dt_no_later_than)]
    ret.sort(reverse=not ascending)

    return ret


def search_files_by_pattern(file_patten, folder=None, exclude_str=None):
    """
    >>> from qtc.ext.path import sysequity_dir
    >>> folder = quant_risk_dir('EPA_Hist', 'Archive', '20160104')
    >>> import qtc.utils.file_system_utils as fsu
    >>> fsu.search_files_by_pattern(file_patten='summary.*.*.pickle.gz', folder=folder)
    """
    if folder is not None:
        file_patten = os.path.join(folder, file_patten)

    files = glob.glob(file_patten)

    return files if exclude_str is None else [file for file in files if exclude_str not in file]


def search_single_file_by_pattern(file_patten, folder=None, log_error=True):
    """
    >>> from qtc.ext.path import sysequity_dir
    >>> folder = sysequity_dir('EPA_Hist', 'Archive', '20160104')
    >>> import qtc.utils.file_system_utils as fsu
    >>> fsu.search_single_file_by_pattern(file_patten='summary_GICS_IndustrySector.*.*.pickle.gz', folder=folder)
    >>> fsu.search_single_file_by_pattern(file_patten='test.*.*.pickle.gz', folder=folder)

    """
    files = search_files_by_pattern(file_patten=file_patten, folder=folder)
    if len(files) != 1:
        if log_error:
            logger.error(f"Expecting exactly ONE file with pattern: {file_patten} in folder={folder}, "
                         f"however found {len(files)} !")
        return None

    return files[0]


def read_yaml_config_recursively(yaml_config_file):
    if yaml_config_file is None:
        return None

    config_main = None
    try:
        with open(yaml_config_file) as fh:
            config_main = yaml.safe_load(fh, Loader=yaml.FullLoader)
    except Exception as e:
        logger.error(f"Failed to load yaml config from {yaml_config_file}:\n{e}")
        return None

    includes = config_main.get('includes', None)
    config_from_includes = None
    if includes is not None and len(includes) > 0:
        config_from_includes = dict()
        for include in includes:
            include_yaml_config_file = include if os.path.exists(include) else os.path.join(os.path.dirname(yaml_config_file), include)
            config_from_include = read_yaml_config_recursively(yaml_config_file=include_yaml_config_file)
            if config_from_include is not None:
                config_from_includes.update(config_from_include)

    config = dict() if config_from_includes is None else config_from_includes

    if config_main is not None:
        config.update(config_main)

    return config



import unittest, os
import pandas as pd
from qtc.file.file_serialization import FileSerializer, FileSerializationProtocol


class TestFileSerialization(unittest.TestCase):
    def test_FileSerializationProtocol(self):
        self.assertEqual(FileSerializationProtocol.suffix('CSV'), 'csv')
        self.assertEqual(FileSerializationProtocol.protocol('.pkl.gz'), FileSerializationProtocol.PKL_GZ)

    def test_compile_filename(self):
        fsd = FileSerializer(filename_prefix='factor_loadings',
                             filename_dt_component_type='DATEID',
                             protocol='PKL.GZ')
        self.assertEqual(fsd.compile_filename(dt=20210929, extra_filename_components='Archive,AXUS4-MH'),
                         'factor_loadings.Archive.AXUS4-MH.20210929.pkl.gz')

        fsd = FileSerializer(filename_prefix='market_data',
                             filename_dt_component_type='DATEID_HYPHEN_TIMESTR',
                             protocol='html')
        self.assertEqual(fsd.compile_filename(dt='20210929-093000', timezone='America/New_York'),
                         'market_data.20210929-093000.html')

        fsd = FileSerializer(filename_prefix='returns',
                             filename_dt_component_type='DATEID_T_TIMESTR_OFFSETSTR',
                             protocol='PICKLE')
        self.assertEqual(fsd.compile_filename(dt='20210929-093000', timezone='America/New_York'),
                         'returns.20210929T093000-0400.pickle')

    def test_serialization_1(self):
        benchmark = 'This is a\nmultiline\ntext for test.'
        fsd = FileSerializer(filename_prefix='test_serialization_1',
                             filename_dt_component_type='NON_DATETIME',
                             protocol='TXT')
        extra_filename_components = 'local,test'
        file_path = fsd.serialize(data=benchmark, folder=None, extra_filename_components=extra_filename_components)
        file_path = os.path.abspath(file_path)
        target_file_path = fsd.search_file(folder=os.path.dirname(file_path), extra_filename_components=extra_filename_components)
        self.assertEqual(target_file_path, file_path)

        target = fsd.deserialize(folder=os.path.dirname(file_path), extra_filename_components=extra_filename_components)
        self.assertEqual(target, benchmark)

        os.remove(file_path)

    def test_serialization_2(self):
        benchmark = pd.DataFrame({'Col1':[1,2,3], 'Col2':['a','b','c']}, index=['Row1','Row2','Row3'])
        fsd = FileSerializer(filename_prefix='test_serialization_2',
                             filename_dt_component_type='DATEID',
                             protocol='PKL_GZ')
        folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'benchmark')
        dt = 20210930
        file_path = fsd.serialize(data=benchmark, folder=folder, dt=dt)
        target_file_path = fsd.search_file(folder=folder, dt=dt)
        self.assertEqual(target_file_path, file_path)

        target = fsd.deserialize(folder=os.path.dirname(file_path), dt=dt)
        pd.testing.assert_frame_equal(target, benchmark)

        os.remove(file_path)

    def test_serialization_3(self):
        benchmark = pd.DataFrame({'Col1':[1,2,3], 'Col2':['a','b','c']}, index=['Row1','Row2','Row3'])
        benchmark.reset_index(inplace=True)
        fsd = FileSerializer(filename_prefix='test_serialization_3',
                             filename_dt_component_type='DATEID',
                             protocol='FEATHER')
        folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'benchmark')
        dt = 20210930
        file_path = fsd.serialize(data=benchmark, folder=folder, dt=dt)
        target_file_path = fsd.search_file(folder=folder, dt=dt)
        self.assertEqual(target_file_path, file_path)

        target = fsd.deserialize(folder=os.path.dirname(file_path), dt=dt)
        pd.testing.assert_frame_equal(target, benchmark)

        os.remove(file_path)

    def test_serialization_4(self):
        benchmark = pd.DataFrame({'Col1':[1,2,3], 'Col2':['a','b','c']}, index=['Row1','Row2','Row3'])
        benchmark.reset_index(inplace=True)
        fsd = FileSerializer(filename_prefix='test_serialization_4',
                             filename_dt_component_type='DATEID',
                             protocol='FTR')
        folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'benchmark')
        dt = 20210930
        file_path = fsd.serialize(data=benchmark, folder=folder, dt=dt)
        target_file_path = fsd.search_file(folder=folder, dt=dt)
        self.assertEqual(target_file_path, file_path)

        target = fsd.deserialize(folder=os.path.dirname(file_path), dt=dt)
        pd.testing.assert_frame_equal(target, benchmark)

        os.remove(file_path)

    def test_serialization_5(self):
        benchmark = pd.DataFrame({'Col1':[1,2,3], 'Col2':['a','b','c']}, index=['Row1','Row2','Row3'])
        benchmark.reset_index(inplace=True)
        fsd = FileSerializer(filename_prefix='test_serialization_5',
                             filename_dt_component_type='DATEID',
                             protocol='JAY')
        folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'benchmark')
        dt = 20220525
        file_path = fsd.serialize(data=benchmark, folder=folder, dt=dt)
        target_file_path = fsd.search_file(folder=folder, dt=dt)
        self.assertEqual(target_file_path, file_path)

        target = fsd.deserialize(folder=os.path.dirname(file_path), dt=dt)
        pd.testing.assert_frame_equal(target, benchmark)

        os.remove(file_path)


if __name__ == '__main__':
    unittest.main()



import qtc.ext.unittest as ut
from qtc.ext.inspect import inspect_caller


class TestInspect(ut.TestCase):
    def test_inspect_caller(self):
        benchmark = {'CLASS': 'TestInspect',
                     'FUNC': 'test_inspect_caller',
                     'LINENO': 11,
                     'MODULE': 'test_inspect'}
        target = inspect_caller(return_info=['MODULE', 'CLASS', 'FUNC', 'LINENO'])
        target['MODULE'] = target['MODULE'].split('.')[-1]
        self.assertEqual(first=benchmark, second=target)


if __name__ == '__main__':
    ut.main()



import qtc.utils.misc_utils as mu
import plotly.graph_objs as go

####################################################################
# !!! It would be better that all "styles" be specified in style.css
#     instead of in this python module
####################################################################

DEFAULT_DASH_TABLE_STYLES = None
def get_dash_table_styles(apply_defaults=True,
                          bold_cols=None):
    """
    >>> import sysequity_common.utils.dash_utils as dashu
    """
    global DEFAULT_DASH_TABLE_STYLES
    if apply_defaults:
        if DEFAULT_DASH_TABLE_STYLES is None:
            style_data_conditional = [{
                'if': {'row_index': 'odd'},
                # 'backgroundColor': 'rgb(220, 220, 220)',
                'backgroundColor': '#D9E1F2',
            }]

            DEFAULT_DASH_TABLE_STYLES = dict(
                css=[{'selector': 'table', 'rule': 'table-layout: fixed'}],
                style_header={
                    'textAlign': 'center',
                    'fontWeight': 'bold',
                    # 'backgroundColor': '#0047AB',
                    'backgroundColor': '#305496',
                    'color': 'white'
                },
                style_data_conditional=style_data_conditional,
                style_cell={'padding': '8px', 'font-family': 'sans-serif', 'font-size': '12px', 'whiteSpace': 'pre-line'}, # 'Monaco'
                fill_width=False
            )

        dash_table_styles = DEFAULT_DASH_TABLE_STYLES
    else:
        dash_table_styles = dict()

    if bold_cols is not None:
        bold_cols = set(mu.iterable_to_tuple(bold_cols, raw_type='str'))
        style_data_conditional = dash_table_styles.get('style_data_conditional', list())
        style_data_conditional += [{'if': {'column_id': bold_col}, 'fontWeight': 'bold'} for bold_col in bold_cols]
        dash_table_styles['style_data_conditional'] = style_data_conditional

    return dash_table_styles


def set_dash_table_new_line(col_fmts, cols_with_new_line):
    new_fmt_items = list()
    for col in cols_with_new_line:
        col_name = col.replace(' ', '\n')
        found = False
        for col_fmt in col_fmts:
            id_value = col_fmt.get('id', None)
            if id_value is not None and id_value == col:
                col_fmt['name'] = col_name
                found = True

        if not found:
            new_fmt_items.append([{'id': col, 'name': col_name}])

    col_fmts.extend(new_fmt_items)

    return col_fmts



DEFAULT_TAB_STYLES = None
def get_default_tab_style():
    global DEFAULT_TAB_STYLES
    if DEFAULT_TAB_STYLES is None:
        DEFAULT_TAB_STYLES = {
            # 'borderBottom': '1px solid #d6d6d6',
            'padding': '6px',
            # 'fontWeight': 'bold',
            'backgroundColor': '#305496',
            'color': 'white',
        }
    return DEFAULT_TAB_STYLES


DEFAULT_TAB_SELECTED_STYLES = None
def get_default_tab_selected_style():
    global DEFAULT_TAB_SELECTED_STYLES
    if DEFAULT_TAB_SELECTED_STYLES is None:
        DEFAULT_TAB_SELECTED_STYLES = {
            # 'borderBottom': '1px solid #d6d6d6',
            'padding': '6px',
            # 'fontWeight': 'bold',
            'backgroundColor': 'white',
            'color': 'black',
        }
    return DEFAULT_TAB_SELECTED_STYLES


DEFAULT_LEGEND_STYLES = None
def get_default_legend_styles():
    global DEFAULT_LEGEND_STYLES
    if DEFAULT_LEGEND_STYLES is None:
        DEFAULT_LEGEND_STYLES = {
            'orientation': 'h',
            'title': None,
            'x': 1, 'xanchor': 'right',
            'y': 1, 'yanchor': 'bottom'}
    return DEFAULT_LEGEND_STYLES


def blank_fig():
    fig = go.Figure(go.Scatter(x=[], y=[]))
    fig.update_layout(template=None)
    fig.update_xaxes(showgrid=False, showticklabels=False, zeroline=False)
    fig.update_yaxes(showgrid=False, showticklabels=False, zeroline=False)

    return fig



import qtc.ext.unittest as ut
import qtc.env_config as ecfg


class TestEnvConfig(ut.TestCase):
    def test_get_env_config(self):
        self.assertEqual(ecfg.get_env_config().environment,
                         'RESEARCH')
        self.assertEqual(ecfg.get_env_config().get(prop='GMBP-RDS.host'),
                         'rm-rj9e2t4ee6xr208gk6o.mysql.rds.aliyuncs.com')


if __name__ == '__main__':
    ut.main()


import inspect
from typing import List


def inspect_caller(skip: int = 1,
                   return_info: List[str] = ('MODULE', 'FILENAME', 'CLASS', 'FUNC', 'LINENO')):
    """Get the name of a caller in the format module.class.method.

    :param skip: Specifies how many levels of stack to skip while getting caller name.
                 skip=1 means "who calls me"; skip=2 means "who calls my caller" etc.
    :param return_info: Caller info keys in returned dictionary: [MODULE | FILENAME | CLASS | FUNC | LINENO]
    :return: Caller info dictionary with keys specified in "return_info",
             while None return indicates empty "return_info" or "skip" exceeds stack height.
    """

    if not return_info:
        return None

    stack = inspect.stack()
    start = 0 + skip
    if len(stack) < start + 1:
        return None

    caller_info = dict()
    parent_frame = stack[start][0]

    # package and module
    if 'MODULE' in return_info:
        module_info = inspect.getmodule(parent_frame)
        if module_info:
            caller_info['MODULE'] = module_info.__name__

    # caller filename
    if 'FILENAME' in return_info:
        caller_info['FILENAME'] = parent_frame.f_code.co_filename

    # class
    if 'CLASS' in return_info:
        klass = None
        if 'self' in parent_frame.f_locals:
            klass = parent_frame.f_locals['self'].__class__.__name__
        caller_info['CLASS'] = klass

    # method or function
    if 'FUNC' in return_info:
        func = None
        if parent_frame.f_code.co_name != '<module>':  # top level usually
            func = parent_frame.f_code.co_name
        caller_info['FUNC'] = func

    # line in caller
    if 'LINENO' in return_info:
        caller_info['LINENO'] = parent_frame.f_lineno

    # Remove reference to frame
    # See: https://docs.python.org/3/library/inspect.html#the-interpreter-stack
    del parent_frame

    return caller_info



import unittest, os
import pandas as pd
from qtc.file.file_cache import FileCache

BENCHMARK = pd.DataFrame({'Col1':[1, 2, 3],
                          'Col2': ['a', 'b', 'c']})

@FileCache
def load_data(dateid, factor_model_code,
              **kwargs):
    data = BENCHMARK
    return data


@FileCache
def load_data_new(date, factor_model_code,
                  file_cache_mode='l', file_cache_dir=None, file_cache_name='FUNC_NAME_DATEID', file_cache_protocol='CSV'):
    data = BENCHMARK
    return data


class TestFileCache(unittest.TestCase):
    def setUp(self) -> None:
        self.dateid = 20211019
        self.factor_model_code = 'BARRA-USFASTD'

    def test_file_cache_mode_disabled(self):
        data = load_data(dateid=self.dateid,
                         factor_model_code=self.factor_model_code,
                         file_cache_mode='DISABLED')
        pd.testing.assert_frame_equal(data, BENCHMARK)

    def test_file_cache_mode_overwrite_enabled(self):
        data = load_data(dateid=self.dateid,
                         factor_model_code=self.factor_model_code,
                         file_cache_mode='OVERWRITE')
        data = load_data(dateid=self.dateid,
                         factor_model_code=self.factor_model_code,
                         file_cache_mode='ENABLED')

        pd.testing.assert_frame_equal(data, BENCHMARK)

        os.remove(f'load_data.dateid={self.dateid}.factor_model_code={self.factor_model_code}.csv')

    def test_file_cache_name_func_name_dateid(self):
        data = load_data(dateid=self.dateid,
                         factor_model_code=self.factor_model_code,
                         file_cache_name='FUNC_NAME_DATEID',
                         file_cache_mode='OVERWRITE')
        data = load_data(dateid=self.dateid,
                         factor_model_code=self.factor_model_code,
                         file_cache_name='FUNC_NAME_DATEID',
                         file_cache_mode='ENABLED')

        pd.testing.assert_frame_equal(data, BENCHMARK)

        os.remove(f'load_data.{self.dateid}.csv')

    def test_file_cache_name_func_name_dateid_1(self):
        data = load_data_new(date=self.dateid,
                             factor_model_code=self.factor_model_code,
                             file_cache_mode='OVERWRITE')
        data = load_data_new(date=str(self.dateid),
                             factor_model_code=self.factor_model_code,
                             file_cache_mode='ENABLED')

        pd.testing.assert_frame_equal(data, BENCHMARK)

        os.remove(os.path.join(os.getcwd(), f'load_data_new.{self.dateid}.csv'))

    def test_file_cache_name_user_defined(self):
        data = load_data(dateid=self.dateid,
                         factor_model_code=self.factor_model_code,
                         file_cache_name='my_test_file_cache',
                         file_cache_mode='OVERWRITE')
        data = load_data(dateid=self.dateid,
                         factor_model_code=self.factor_model_code,
                         file_cache_name='my_test_file_cache',
                         file_cache_mode='ENABLED')

        pd.testing.assert_frame_equal(data, BENCHMARK)

        os.remove(f'my_test_file_cache.csv')

    def test_file_cache_protocol_pkl(self):
        data = load_data(dateid=self.dateid,
                         factor_model_code=self.factor_model_code,
                         file_cache_name='my_test_file_cache',
                         file_cache_protocol='PKL',
                         file_cache_mode='OVERWRITE')
        data = load_data(dateid=self.dateid,
                         factor_model_code=self.factor_model_code,
                         file_cache_name='my_test_file_cache',
                         file_cache_protocol='PKL',
                         file_cache_mode='ENABLED')

        pd.testing.assert_frame_equal(data, BENCHMARK)

        os.remove(f'my_test_file_cache.pkl')

    def test_file_cache_protocol_pkl_gz(self):
        data = load_data(dateid=self.dateid,
                         factor_model_code=self.factor_model_code,
                         file_cache_name='my_test_file_cache',
                         file_cache_protocol='PKL_GZ',
                         file_cache_mode='OVERWRITE')
        data = load_data(dateid=self.dateid,
                         factor_model_code=self.factor_model_code,
                         file_cache_name='my_test_file_cache',
                         file_cache_protocol='PKL_GZ',
                         file_cache_mode='ENABLED')

        pd.testing.assert_frame_equal(data, BENCHMARK)

        os.remove(f'my_test_file_cache.pkl.gz')


if __name__ == '__main__':
    unittest.main()



import json
import pickle
import time
import requests
import platform
import traceback
import pandas as pd
from qtc.consts.enums import DataTransmissionProtocol
from qtc.ext.logging import set_logger
logger = set_logger()


def get_auth(username=None, password=None):
    if username is not None or password is not None:
        from requests_ntlm import HttpNtlmAuth
        auth = HttpNtlmAuth(username=username, password=password)
    else:
        if platform.system() == 'Windows':
            from requests_negotiate_sspi import HttpNegotiateAuth
            return HttpNegotiateAuth()
        else:
            from requests_kerberos import HTTPKerberosAuth, OPTIONAL
            return HTTPKerberosAuth(mutual_authentication=OPTIONAL)

    return auth


def _process_request(request_url, is_post, body=None,
                     username=None, password=None,
                     data_transmission_protocol='JSON'):
    # import time
    # start_time = time.time()

    with requests.Session() as session:
        session.auth = get_auth(username=username, password=password)

        if is_post:
            response = session.post(request_url, json=body, timeout=300)
        else:
            response = session.get(request_url, params=body, timeout=300)

        # print(f"Request roundtrip time={time.time()-start_time} @ request_url={request_url}")
        # start_time = time.time()

        if not response.ok:
            url_format = lambda x: x if len(x)<100 else x[0:100]+'...'
            logger.error(f'Failed with status code {response.status_code} for '
                         f'URL: {url_format(request_url)} with Body: {url_format(str(body))}')
            response.raise_for_status()
            return None

        if data_transmission_protocol is None:
            return response

        if isinstance(data_transmission_protocol, str):
            data_transmission_protocol = DataTransmissionProtocol.retrieve(data_transmission_protocol)

        if data_transmission_protocol==DataTransmissionProtocol.JSON:
            data = json.loads(response.content.decode('utf-8'))
            # print(f"JSON parsing time = {time.time()-start_time} @ request_url={request_url}")
        elif data_transmission_protocol==DataTransmissionProtocol.PICKLE:
            data = pickle.loads(response.content)
            # print(f"pickle.loads elapse time = {time.time()-start_time} @ request_url={request_url}")
        elif data_transmission_protocol==DataTransmissionProtocol.TXT:
            data = response.content.decode('utf-8')

    return data


def process_request(request_url, is_post, body=None,
                    username=None, password=None,
                    data_transmission_protocol='JSON',
                    max_retry=0, retry_interval=60):
    data = None
    n_try = 0
    while data is None:
        try:
            data = _process_request(request_url=request_url, is_post=is_post, body=body,
                                    username=username, password=password,
                                    data_transmission_protocol=data_transmission_protocol)
        except:
            traceback.print_exc()

        n_try += 1
        if n_try>max_retry:
            break

        logger.warn(f'Sleeping for {retry_interval} seconds before {n_try} retry ...')
        time.sleep(retry_interval)

    if data is None and max_retry>0:
        logger.error(f'Failed to get data after {max_retry} retries !')

    return data


def process_request_json2df(request_url, is_post, body=None,
                            username=None, password=None,
                            max_retry=0, retry_interval=60):
    data = process_request(request_url=request_url, is_post=is_post, body=body,
                           username=username, password=password,
                           data_transmission_protocol=DataTransmissionProtocol.JSON,
                           max_retry=max_retry, retry_interval=retry_interval)
    if data is None:
        return None

    try:
        data = pd.DataFrame.from_records(data['Data']['Records'],
                                         columns=data['Data']['Fields'])
    except:
        try:
            data = pd.DataFrame(data['Data'])
        except:
            try:
                data = pd.DataFrame(data)
            except:
                logger.error(f"Failed to parse JSON data to pd.DataFrame!\n{data}")

    return data



import os
import re
import time
import getpass
import traceback
import pandas as pd
import sqlalchemy
from urllib.parse import quote
from typing import Optional, Dict
from qtc.consts.enums import DBType
import qtc.env_config as ecfg
import qtc.utils.misc_utils as mu
from qtc.ext.logging import set_logger
logger = set_logger()

import warnings
warnings.filterwarnings('ignore')


def _get_conn_mssql(host, port, user, password,
                    database=None):
    if port is None:
        port = 1433

    logger.debug(f'host={host}, port={port}, user={user}, password={password}, database={database}')
    import pymssql
    if os.name == 'nt':  # windows
        conn = pymssql.connect(user=user, password=password,
                               host=host, port=port)
    elif os.name == 'posix':  # linux
        if database:
            conn = pymssql.connect(user=user, password=password,
                                   host=host, port=port,
                                   database=database)
        else:
            conn = pymssql.connect(user=user, password=password,
                                   host=host, port=port)

    if conn:
        conn.autocommit(True)

    return conn


def _get_conn_redshift(host, port, user, password,
                       database=None):
    # if host.upper() == 'REDSHIFT':
    #     host = ecfg.get_env_config().get(prop='redshift.host')
    #     if port is None:
    #         port = ecfg.get_env_config().get(prop='redshift.port', default=5439)
    if port is None:
        port = ecfg.get_env_config().get(prop='redshift.port', default=5439)

    if database is None:
        database = 'bigdataprod'

    logger.debug(f'host={host}, port={port}, user={user}, password={password}, database={database}')
    import psycopg2
    conn = psycopg2.connect(user=user, password=password,
                            host=host, port=port,
                            database=database)
    if conn:
        conn.set_session(autocommit=True)

    return conn


def _get_engine(db_type,
                host, port, user, password,
                database=None):
    if isinstance(db_type, str):
        db_type = DBType.retrieve(value=db_type)

    if db_type==DBType.POSTGRES:
        conn_string = f'postgresql://{user}:{quote(password)}@{host}:{port}'
    else:
        raise Exception(f'db_type={db_type} is not supported in [POSTGRES] !')

    if database is not None:
        conn_string = f'{conn_string}/{database}'

    engine = sqlalchemy.create_engine(conn_string)
    return engine


def _get_conn_postgres(host, port, user, password,
                       database=None):
    engine = _get_engine(db_type=DBType.POSTGRES,
                         host=host, port=port, user=user, password=password,
                         database=database)
    conn = engine.connect()
    return conn


def _get_conn_mysql(host, port, user, password,
                    database=None):
    # engine = _get_engine(db_type=DBType.MYSQL,
    #                      host=host, port=port, user=user, password=password,
    #                      database=database)
    # conn = engine.connect()
    import pymysql
    conn = pymysql.connect(host=host, port=int(port), user=user, passwd=password, db=database)

    return conn


def _init_conn(host, port=None,
               user=None, password=None,
               database=None,
               db_type=None):

    conn = None
    if db_type is None:
        try:
            conn = _get_conn_mysql(host=host, port=port, user=user, password=password, database=database)
        except:
            try:
                conn = _get_conn_mssql(host=host, port=port, user=user, password=password, database=database)
            except:
                try:
                    conn = _get_conn_postgres(host=host, port=port, user=user, password=password, database=database)
                except:
                    try:
                        conn = _get_conn_redshift(host=host, port=port, user=user, password=password, database=database)
                    except:
                        logger.warn(f'Unable to create connection by db_type=[MSSQL|MYSQL|REDSHIFT|POSTGRES] with '
                                    f'host={host}, port={port}, user={user}, password={password}, database={database} !')
                        return None
        return conn

    if isinstance(db_type, str):
        db_type = DBType.retrieve(value=db_type)

    if user is None:
        user = getpass.getuser()

    if db_type==DBType.MSSQL:
        conn = _get_conn_mssql(host=host, port=port, user=user, password=password, database=database)
    elif db_type==DBType.MYSQL:
        conn = _get_conn_mysql(host=host, port=port, user=user, password=password, database=database)
    elif db_type==DBType.REDSHIFT:
        conn = _get_conn_redshift(host=host, port=port, user=user, password=password, database=database)
    elif db_type==DBType.POSTGRES:
        conn = _get_conn_postgres(host=host, port=port, user=user, password=password, database=database)
    else:
        logger.error(f'db_type={db_type} is not supported in [MSSQL | REDSHIFT | POSTGRES] !')

    return conn


def _get_conn_with_retry(host, port=None,
                         user=None, password=None,
                         database=None,
                         db_type=None,
                         max_retry=0, retry_interval=60):
    conn = None
    n_try = 0
    while conn is None and n_try<=max_retry:
        if n_try != 0:
            logger.warn(f'Sleeping for {retry_interval} seconds before {n_try} retry ...')
            time.sleep(retry_interval)

        try:
            conn = _init_conn(host=host, port=port,
                              user=user, password=password,
                              database=database,
                              db_type=db_type)
        except:
            pass

        n_try += 1

    if conn is None and max_retry>0:
        logger.error(f'Failed to init conn after {max_retry} retries !')

    return conn


def _infer_db_type_name(db_code):
    """
    >>> import qtc.utils.db_utils as dbu
    >>> dbu._infer_db_type_name(db_code='DW-DB')
    (<DBType.MSSQL: 'MSSQL'>, 'DW-DB')
    >>> dbu._infer_db_type_name(db_code='DW-DB.host')
    (<DBType.MSSQL: 'MSSQL'>, 'DW-DB')
    >>> dbu._infer_db_type_name(db_code='redshift.host')
    (<DBType.REDSHIFT: 'REDSHIFT'>, 'redshift')
    >>> dbu._infer_db_type_name(db_code='redshift.bamfunds.net')
    (<DBType.REDSHIFT: 'REDSHIFT'>, 'redshift.bamfunds.net')
    """
    db_name = db_code.replace('.host', '') if '.host' in db_code else db_code

    mssql_dbs = ecfg.get_env_config().get('mssql_dbs', '')
    mssql_dbs = set(mu.iterable_to_tuple(mssql_dbs, raw_type='str'))

    mysql_dbs = ecfg.get_env_config().get('mysql_dbs', 'GMBP-ADS,GMBP-RDS')
    mysql_dbs = set(mu.iterable_to_tuple(mysql_dbs, raw_type='str'))

    postgres_dbs = ecfg.get_env_config().get('postgres_dbs', '')
    postgres_dbs = set(mu.iterable_to_tuple(postgres_dbs, raw_type='str'))

    if db_name.upper() in mysql_dbs or db_name.split('.')[0].upper() in mysql_dbs:
        db_type = DBType.MYSQL
    elif db_name.upper() in mssql_dbs or db_name.split('.')[0].upper() in mssql_dbs:
        db_type = DBType.MSSQL
    elif db_name.upper() in postgres_dbs or db_name.split('.')[0].upper() in postgres_dbs:
        db_type = DBType.POSTGRES
    elif db_name.upper() in ('REDSHIFT') or 'redshift' in db_code:
        db_type = DBType.REDSHIFT
    else:
        db_type = None

    return db_type, db_name


def _get_db_config(db_code):
    db_type, db_name = _infer_db_type_name(db_code=db_code)
    host, port, user, password, database = None, None, None, None, None

    if db_type is not None:
        user = ecfg.get_env_config().get(f'{db_name}.user',
                                         os.getenv(f'{db_name}.user',
                                                   os.getenv(f'{db_type.value}_USER', None)))
        password = ecfg.get_env_config().get(f'{db_name}.password',
                                             os.getenv(f'{db_name}.password',
                                                       os.getenv(f'{db_type.value}_PASSWORD', None)))
        port = ecfg.get_env_config().get(f'{db_name}.port', None)

        host = ecfg.get_env_config().get(f'{db_name}.host',
                                         ecfg.get_env_config().get(db_code,
                                                                   os.getenv(f'{db_name}.host',
                                                                             os.getenv(f'{db_type.value}_HOST', db_code))))
        if db_type==DBType.POSTGRES:
            database = db_name

        # if db_type == DBType.REDSHIFT:
        #     host = ecfg.get_env_config().get(prop='{}.host')
        #     if port is None:
        #         port = ecfg.get_env_config().get(prop='redshift.port', default=5439)
    else:
        host = db_code

    return host, port, user, password, database, db_type, db_name


def get_conn(db_code):
    host, port, user, password, database, db_type, db_name = _get_db_config(db_code=db_code)

    return _get_conn_with_retry(host=host, port=port,
                                user=user, password=password,
                                database=database,
                                db_type=db_type,
                                max_retry=0, retry_interval=60)


def execute_sql(sql,
                host=None, conn=None,
                log_sql=False):
    if conn is None:
        conn = get_conn(db_code=host)

    if conn is None:
        logger.error(f'Failed to get connection to run sql={sql} against host={host} !')

    if log_sql:
        logger.debug(f"Connection: {conn}")
        logger.debug(f"SQL Query:\n{sql}")

    if os.path.isfile(sql):
        with open(sql, 'r') as fh:
            sql = fh.read()

    try:
        cursor = conn.cursor()
        cursor.execute(sql)
    except Exception as e:
        logger.error(f"Exception caught when executing:\n{sql}")
        traceback.print_exc()
    finally:
        conn.close()

    return


def read_sql(sql,
             host=None, conn=None,
             log_sql=False, params=None):
    """Wraps pandas read_sql function. For raw SQL queries without columns selections, this function is perferred.

    >>> import qtc.utils.db_utils as dbu
    >>> dbu.read_sql(sql="SELECT * FROM ANTH_REFERENCE.trd.RESTRICTIONS (NOLOCK) " \
                          "WHERE DATA_DATE='2022-05-04'",
                     host='ANTHSYSEQ-DB').head()
        DATA_DATE      CODE  TICKER COUNTRY  TYPE MODE SOURCE           REASON              ENTRY_TIME
    0  2022-05-04  64261093  600764      CN  HOLD  AUT    BAM  BAM Restriction 2022-05-04 05:34:51.330
    1  2022-05-04  64296926  688981      CN  HOLD  AUT    BAM  BAM Restriction 2022-05-04 05:34:51.330
    2  2022-05-04  64366361  003009      CN  HOLD  AUT    BAM  BAM Restriction 2022-05-04 05:34:51.330
    3  2022-05-04  64618032  003009      CN  HOLD  AUT    BAM  BAM Restriction 2022-05-04 05:34:51.330
    4  2022-05-04  64680269  601728      CN  HOLD  AUT    BAM  BAM Restriction 2022-05-04 05:34:51.330
    """

    if conn is None:
        conn = get_conn(db_code=host)

    if conn is None:
        logger.error(f'Failed to get connection to run sql={sql} against host={host} !')

    if log_sql:
        logger.debug(f"Connection: {conn}")
        logger.debug(f"SQL Query:\n{sql}")

    df = None
    try:
        df = pd.read_sql(sql, conn, params=params)
    finally:
        conn.close()

    return df


def _query_headers(sql,
                   host=None, conn=None):
    """Helper function to get headers from DB table contained in the SQL query.
    >>> import qtc.utils.db_utils as dbu
    >>> dbu._query_headers(sql="SELECT * FROM ANTH_REFERENCE.trd.RESTRICTIONS (NOLOCK) " \
                               "WHERE DATA_DATE='2022-05-04'",
                           host='ANTHSYSEQ-DB')
    Index(['DATA_DATE', 'CODE', 'TICKER', 'COUNTRY', 'TYPE', 'MODE', 'SOURCE', 'REASON', 'ENTRY_TIME'], dtype='object')
    """

    try:
        select_top_1_sql = re.sub('SELECT[ ]+.*\s*FROM', 'SELECT TOP 1 * FROM', sql, flags=re.IGNORECASE)
        ret = read_sql(sql=select_top_1_sql, host=host, conn=conn).columns
    except:
        select_top_1_sql = re.sub('SELECT[ ]+.*\s*FROM', 'SELECT * FROM', sql, flags=re.IGNORECASE) + ' FETCH FIRST ROW ONLY'
        ret = read_sql(sql=select_top_1_sql, host=host, conn=conn).columns

    return ret


def _process_cols(available_cols, required_cols=None, expected_cols=None,
                  raise_exception_on_not_found_required_cols=False,
                  sql=None):
    """Infers the final columns based on all "available_cols", "required_cols"(usually key columns in the DB table)
    and user "expected_cols", which could overlap with "required_cols".
    If "required_cols" is not None, then the returned cols will have "required_cols" first and then other columns if more columns should be returned.
    If "expected_cols" is None, then return all "available_cols". Otherwise, return "required_cols"+"expected_cols".

    :param available_cols: All available columns. Please note that it cannot be None.
    :type available_cols: str | int | Iterable. See :func:`sysequity_common.utils.misc.iterable_to_tuple`
    :param required_cols: Required columns. If not found in the available columns, exception handling will depend on 'raise_exception_on_not_found_required_cols'.
    :type required_cols: str | int | Iterable. See :func:`sysequity_common.utils.misc.iterable_to_tuple`
    :param expected_cols: User expected columns.
    :type expected_cols: str | int | Iterable. See :func:`sysequity_common.utils.misc.iterable_to_tuple`
    :param raise_exception_on_not_found_required_cols: If True, throw exception if there are some required columns which cannot be found in available columns.
    :type raise_exception_on_not_found_required_cols: bool
    :return: list - Return columns.

    >>> import qtc.utils.db_utils as dbu
    >>> dbu._process_cols(available_cols='SecurityId,AxiomaId,BbergCode,NMV,Return', required_cols='SecurityId', expected_cols='NMV,Return')
    ['SecurityId', 'NMV', 'Return']
    >>> dbu._process_cols(available_cols='SecurityId,AxiomaId,BbergCode,NMV,Return', required_cols='SecurityId,NMV', expected_cols='NMV,Return,SecurityId')
    ['SecurityId', 'NMV', 'Return']
    >>> dbu._process_cols(available_cols='SecurityId,AxiomaId,BbergCode,NMV,Return', required_cols='SecurityId,NMV,BbergCode')
    ['SecurityId', 'NMV', 'BbergCode', 'AxiomaId', 'Return']
    >>> dbu._process_cols(available_cols='SecurityId,AxiomaId,BbergCode,NMV,Return', required_cols='SecurityId,DisplayCode')
    [2019-12-06 17:24:20,822: bamrisk.utils.db_utils - process_cols()] ERROR - Failed to find these required columns: ['DisplayCode'] in available columns: ['SecurityId', 'AxiomaId', 'BbergCode', 'NMV', 'Return']
    ['SecurityId', 'AxiomaId', 'BbergCode', 'NMV', 'Return']
    >>> dbu._process_cols(available_cols='SecurityId,AxiomaId,BbergCode,NMV,Return', expected_cols='SecurityId,NMV')
    ['SecurityId', 'NMV']
    """
    available_cols = list(mu.iterable_to_tuple(available_cols, raw_type='str'))
    available_cols_upper = [col.upper() for col in available_cols]

    if required_cols is None:
        required_cols = expected_cols

    sql = '' if sql is None else f"\n{sql}"

    ret_cols = list()
    if required_cols is not None:
        # Check if all required_cols are in available_cols
        required_cols = list(mu.iterable_to_tuple(required_cols, raw_type='str'))
        # not_found_cols = [col for col in required_cols if col not in available_cols]
        not_found_cols = [col for col in required_cols if col.upper() not in available_cols_upper]
        if len(not_found_cols) > 0:
            if raise_exception_on_not_found_required_cols:
                raise Exception(f"Failed to find these required columns: {not_found_cols} in available columns: {available_cols}{sql}")
            else:
                logger.error(f"Failed to find these required columns: {not_found_cols} in available columns: {available_cols}{sql}")
                required_cols = [col for col in required_cols if col.upper() in available_cols_upper]

        ret_cols = required_cols

    ret_cols_upper = [col.upper() for col in ret_cols]
    if expected_cols is None:
        for col in available_cols:
            if col.upper() not in ret_cols_upper:
                ret_cols.append(col)

        return ret_cols

    ret_cols_upper = [col.upper() for col in ret_cols]
    expected_cols = list(mu.iterable_to_tuple(expected_cols, raw_type='str'))
    not_found_cols = list()
    for col in expected_cols:
        if col.upper() not in available_cols_upper:
            not_found_cols.append(col)
        else:
            if col.upper() not in ret_cols_upper:
                ret_cols.append(col)

    if len(not_found_cols)>0:
        logger.warn(f"Failed to find these expected columns: {not_found_cols} in available columns: {available_cols}\n{sql}")

    return ret_cols


def _modify_sql(sql, required_cols=None, expected_cols=None,
                available_cols=None,
                host=None, conn=None,
                quotes_on_colnames=False):
    """A helper function to infer the minimal set of columns to query and modify the original query accordingly.

    .. note::
        #. Andrew Hu's philosophy for DB access layer is that the SQL queries should be as "atomic" as possible. So here the "sql" is simple query which usually only queries one DB table.
        #. 'available_cols', 'required_cols', 'expected_cols' are used in :func:`bamrisk_common.utils.db_utils._process_cols` to infer the minimal set of columns.

    :param sql: The SQL query statement.
    :param available_cols: All available columns. If None, then :func:`bamrisk.utils.db_utils.query_table_cols` is called to get the table columns, where 'db_prefix' matters.
    :type available_cols: str | int | Iterable. See :func:`bamrisk.utils.misc.iterable_to_tuple`
    :param required_cols: Required columns. If not found in the available columns, exception handling will depend on 'raise_exception_on_not_found_required_cols'.
    :type required_cols: str | int | Iterable. See :func:`bamrisk.utils.misc.iterable_to_tuple`
    :param expected_cols: User expected columns.
    :type expected_cols: str | int | Iterable. See :func:`bamrisk.utils.misc.iterable_to_tuple`
    :return: str - Modified query with the minimal set of columns in "SELECT".

    >>> import qtc.utils.db_utils as dbu
    >>> modified_sql, cols = dbu._modify_sql(sql='SELECT * FROM BamCorelite.sm.Security WITH (NOLOCK)',
                                             host='DW-DB',
                                             required_cols='SecurityId,Axioma',
                                             expected_cols='SecurityId,Axioma,DisplayCode')
    >>> modified_sql
    'SELECT SecurityId,Axioma,DisplayCode FROM BamCorelite.sm.Security WITH (NOLOCK)'
    >>> cols
    ['SecurityId', 'Axioma', 'DisplayCode']
    """
    if available_cols is None:
        available_cols = _query_headers(sql=sql, host=host, conn=conn)

    cols = _process_cols(available_cols=available_cols, required_cols=required_cols, expected_cols=expected_cols,
                         sql=sql)

    replace = f'''"{'","'.join(cols)}"''' if quotes_on_colnames else ','.join(cols)
    modified_sql = re.sub('SELECT[ ]+.*\s*FROM', f"SELECT {replace} FROM", sql, flags=re.IGNORECASE)

    return modified_sql, cols


def compile_sql_where_clause(sql_and_clauses=None):
    """
    >>> import qtc.utils.db_utils as dbu
    >>> dbu.compile_sql_where_clause()
    ''
    >>> dbu.compile_sql_where_clause("CountryCode IN ('US')")
    "WHERE CountryCode IN ('US')"
    >>> dbu.compile_sql_where_clause(['SecurityId IN (407,409)', 'FactorModelId=60'])
    'WHERE SecurityId IN (407,409)\n\t\tAND FactorModelId=60'
    >>> dbu.compile_sql_where_clause("EffectiveStartDate<='2021-09-23'||EffectiveEndDate>='2020-09-23'")
    "WHERE EffectiveStartDate<='2021-09-23'\n\t\tAND EffectiveEndDate>='2020-09-23'"
    >>> dbu.compile_sql_where_clause(['SecurityId IN (407,409)', 'FactorModelId=60', ''])
    'WHERE SecurityId IN (407,409)\n\t\tAND FactorModelId=60'
    >>> dbu.compile_sql_where_clause([])

    """
    if sql_and_clauses is None or not sql_and_clauses:
        return ''

    sql_and_clauses = [sql_and_clause
                       for sql_and_clause in mu.iterable_to_tuple(sql_and_clauses, raw_type='str', sep='||')
                       if sql_and_clause is not None and sql_and_clause.strip()!='']
    sep = '\n\t\tAND '
    return f'WHERE {sep.join(sql_and_clauses)}' if sql_and_clauses else ''


def sql2df(sql, host=None, conn=None,
           required_cols=None, cols=None,
           preprocess_cols=True,
           log_df_head=False,
           quotes_on_colnames=False):
    """Queries DB server and return results in DataFrame.

    .. note::
        #. It is OK that columns defined in 'cols' are in 'required_cols'. A "union" of these two sets of columns will be returned.
        #. The orders of the columns are the same as in 'required_cols' and then the same as in 'cols'.
        #. If 'cols' is None, all columns will be returned.

    :param sql: SQL query statement.
    :type sql: str
    :param required_cols: Required columns to be in the results. If not found in the results, ERROR will be logged.
    :type required_cols: str | int | Iterable. See :func:`bamrisk.utils.misc.iterable_to_tuple`
    :param cols: Additional columns to keep in the results. If not found in the results, WARNING will be logged.
    :type cols: str | int | Iterable. See :func:`bamrisk.utils.misc.iterable_to_tuple`
    :return: DataFrame - The query results.

    >>> import qtc.utils.db_utils as dbu
    >>> dbu.sql2df(sql="SELECT * FROM ANTH_REFERENCE.ref.SECURITIES_MASTER_ARIES (NOLOCK) " \
                       "WHERE TICKER IN ('TSLA','GOOG') AND DATA_DATE='2022-05-04'",
                   host='ANTHSYSEQ-DB',
                   cols='DATA_DATE,CODE,SID,TICKER,EXCHANGE,BBG_FIGI,BBG_CFIGI')
        DATA_DATE      CODE           SID TICKER EXCHANGE      BBG_FIGI     BBG_CFIGI
    0  2022-05-04  10963872  ARS5c0d0abd8   TSLA      NAS  BBG000N9MNX3  BBG000N9MNX3
    1  2022-05-04  11131355  ARSbb22a4d90   GOOG      NAS  BBG009S3NB30  BBG009S3NB30
    2  2022-05-04  41228590  ARSefe51e330   GOOG      NLB  BBG0123139Y8  BBG012313950
    3  2022-05-04  41228591  ARS091d9b002   TSLA      NLB  BBG012313HF1  BBG012313GN4
    4  2022-05-04  62539930  ARS113f8d1a5   GOOG      SGO  BBG00YJL3ZW2  BBG00YJL3ZV3
    >>> dbu.sql2df(sql="SELECT * FROM ANTH_REFERENCE.ref.SECURITIES_MASTER_ARIES (NOLOCK) " \
                       "WHERE DATA_DATE='2022-05-04'",
                   host='ANTHSYSEQ-DB',
                   required_cols='DATA_DATE,CODE').head()
        DATA_DATE  CODE           SID   SID_MKTDATA     ISSUER_ID TICKER  PRI_RIC CONSOL_RIC MIC_CODE EXCHANGE COUNTRY CURRENCY SECTYPE SEC_SUBTYPE       SEC_DESC  PRIM_LIST_FLAG  COMPOSITE_FLAG COUNTRY_ORIGIN COUNTRY_EXPOSURE COUNTRY_INCORPORATION REGION    SEDOL      CUSIP          ISIN        CIK  CIQ_TRADINGITEMID  CIQ_COMPANYID   GVKEY IID BBG_TICKER_EXCH      BBG_FIGI     BBG_CFIGI    QAI_ID QAI_REGION  QAI_SEC_CODE  DS_INFO_CODE  DS_EXCH_INT_CODE FSYM_REGIONAL_ID FSYM_SECURITY_ID FSYM_ENTITY_ID  AXIOMA_ID  BAM_SECURITY_ID BAM_GENEVA_ID  BAM_MARKIT IBES_EST_PERM_ID IBES_TICKER IDC_CODE  WSPIT_SEC_CODE RIC_ARIES EXCHANGE_ARIES          ENTRY_TIME
    0  2022-05-04   779  ARS16ae50919  ARS16ae50919  ARC2b35627d7   AAON  AAON.OQ     AAON.O     XNGS      NAS      US      USD      EQ          cs    COM USD.001           False            True             US               US                    US   namr  2268130  000360206  US0003602069   824142.0          2585893.0       320500.0  021542  01         AAON US  BBG000C2LZP3  BBG000C2LZP3      AAON          U           779         52508               135         NWCBDX-R         HSDCT7-S       000ZGQ-E  6HS7Y1UZ3           2652.0             0     36279.0      30064826811        AAON  U-43029         30428.0    AAON.O            NAS 2022-05-04 04:59:57
    1  2022-05-04   780  ARS99baa08ba  ARS99baa08ba  ARCc494a810f    AIR    AIR.N        AIR     XNYS      NYS      US      USD      EQ          cs       COM USD1           False            True             US               US                    US   namr  2001119  000361105  US0003611052     1750.0          2585895.0       168154.0  001004  01          AIR US  BBG000BBGQF1  BBG000BBGQF1       AIR          U           780         72943               145         QG1FM5-R         WHQNFK-S       000BFS-E  H2SHPJHH8           2919.0             0     17747.0      30064827829         AIR  U-43477         50933.0       AIR            NYS 2022-05-04 04:59:57
    2  2022-05-04   784  ARS36894bbca  ARS36894bbca  ARCe6f2a8c15    ABB    ABB.N        ABB     XNYS      NYS      US      USD     ADR         adr  SPONSORED ADR           False            True             CH               CH                    CH   namr  2748472  000375204  US0003752047  1091587.0          2585899.0        18527.0  210418  91          ABB US  BBG000DK5Q25  BBG000DK5Q25  00037510          U           784         52570               145         G9ZKT1-R         PRM2JP-S       0036X2-E  TV4GK95L0           2676.0             0      3807.0      30064830118        BBOV    U-903         51405.0       ABB            NYS 2022-05-04 04:59:57
    3  2022-05-04   829  ARS4d332fd30  ARS4d332fd30  ARC2c0172027   ATSG  ATSG.OQ     ATSG.O     XNGS      NAS      US      USD      EQ          cs            COM           False            True             US               US                    US   namr  2938002  00922R105  US00922R1059   894081.0         44806861.0      3545047.0  155280  01         ATSG US  BBG000BLW5F7  BBG000BLW5F7     ABXAV          U           829         52542               135         L7HH56-R         SGHZ0K-S       071X8W-E  SMHT446B9           3288.0             0     51460.0      30064826978         ABX  U-75904          6091.0    ATSG.O            NAS 2022-05-04 04:59:57
    4  2022-05-04   835  ARS1f856790b  ARS1f856790b  ARCfad8ed944   ACCO   ACCO.N     ACCO.K     XNYS      NYS      US      USD      EQ          cs            COM           False            True             US               US                    US   namr  B0G7SZ5  00081T108  US00081T1088   712034.0         23428253.0      4184604.0  164506  02         ACCO US  BBG000J06K07  BBG000J06K07     ABDWI          U           835         70217               145         RPWCT1-R         R2KQ06-S       066M40-E  DPXVZQ4D0           2686.0             0    131582.0      30064826893        ABDW  U-96808         29276.0    ACCO.K            NYS 2022-05-04 04:59:57
    """

    df = pd.DataFrame()
    if preprocess_cols:
        modified_sql, ret_cols = _modify_sql(sql=sql, host=host, conn=conn,
                                             required_cols=required_cols, expected_cols=cols,
                                             quotes_on_colnames=quotes_on_colnames)
        if ret_cols:
            df = read_sql(sql=modified_sql, host=host, conn=conn)
    else:
        df = read_sql(sql=sql, host=host, conn=conn)
        ret_cols = _process_cols(available_cols=df.columns, required_cols=required_cols, expected_cols=cols,
                                 sql=sql)
        if ret_cols:
            df = df[ret_cols]

    if log_df_head:
        logger.info(f"Dataframe with shape {df.shape} returned from\n{sql}\nExamples:\n{df.head()}")

    return df


def create_upsert_method(db_code=None, schema=None,
                         meta: sqlalchemy.MetaData = None, extra_update_fields: Optional[Dict[str, str]] = None):
    """
    Create upsert method that satisfied the pandas's to_sql API.
    """

    if meta is None:
        if db_code is None:
            raise Exception(f'Please provide either "meta" or "db_code" !')

        host, port, user, password, database, db_type, db_name = _get_db_config(db_code=db_code)
        if db_type==DBType.POSTGRES:
            db_engine = _get_engine(db_type=db_type,
                                    host=host, port=port, user=user, password=password,
                                    database=database)
            meta = sqlalchemy.MetaData(db_engine)
            if schema is not None:
                meta.schema = schema

    def method(table, conn, keys, data_iter):
        # select table that data is being inserted to (from pandas's context)
        sql_table = sqlalchemy.Table(table.name, meta, autoload=True)

        # list of dictionaries {col_name: value} of data to insert
        values_to_insert = [dict(zip(keys, data)) for data in data_iter]

        # create insert statement using postgresql dialect.
        # For other dialects, please refer to https://docs.sqlalchemy.org/en/14/dialects/

        if db_type==DBType.POSTGRES:
            insert_stmt = sqlalchemy.dialects.postgresql.insert(sql_table, values_to_insert)
        else:
            raise Exception(f'db_type={db_type} is not supported in [POSTGRES] !')

        # create update statement for excluded fields on conflict
        update_stmt = {exc_k.key: exc_k for exc_k in insert_stmt.excluded}
        if extra_update_fields:
            update_stmt.update(extra_update_fields)

        # create upsert statement.
        upsert_stmt = insert_stmt.on_conflict_do_update(
            index_elements=sql_table.primary_key.columns,  # index elements are primary keys of a table
            set_=update_stmt  # the SET part of an INSERT statement
        )

        # execute upsert statement
        conn.execute(upsert_stmt)

    return method


"""
This module contains some tools to create cleaner report-like output in python notebooks
"""

import base64, io
import matplotlib.pyplot as plt
import nbformat
from IPython.core.display import HTML, Markdown, display, Javascript
from bs4 import BeautifulSoup
from jinja2 import Template
from matplotlib.axes._subplots import SubplotBase
from nbconvert import HTMLExporter, PDFExporter
from nbconvert.preprocessors import ExecutePreprocessor, TagRemovePreprocessor
from qtc.utils.misc_utils import get_zero_centered_min_max
import pandas as pd
import seaborn as sns
sns.set()


def fig_to_base64(fig: SubplotBase):
    """
    Encodes figures for placing directly into html
    :param fig: figure to be encoded
    :return: base64 encoded plot
    """
    img = io.BytesIO()
    fig.savefig(img, format='png', bbox_inches='tight')
    img.seek(0)
    return base64.b64encode(img.getvalue())


def output_grid(*functions, nrows=1, ncols=1, byrow=True):
    """
    Input any number of functions (() ->  obj) that resolve to an object
    which either has _repr_html_ or are a matplotlib SubPlotBase and generate
    a grid display of the results in html.
    :param functions: Any number of () -> obj functions
    :param nrows: number of rows in the grid
    :param ncols: number of columns in the grid
    :param byrow: functions should be taken in row major order (fill row1, then row2, etc).
                  False is column majore order.  Rows are filled left to right fro the top,
                  Columns are filled top to bottom from the left
    :return: html
    """
    # Get the html outputs
    html_outputs = []
    for function in functions:
        obj = function() if callable(function) else function
        if isinstance(obj, SubplotBase):
            fig = obj.get_figure()
            encoded = fig_to_base64(fig)
            plt.close(fig)
            html_outputs.append('<img src="data:image/png;base64, {}">'.format(encoded.decode('utf-8')))
        elif hasattr(obj, '_repr_html_') and callable(obj._repr_html_):
            html_outputs.append(obj._repr_html_())
        elif isinstance(obj, Markdown):
            html_outputs.append(obj)
        else:
            html_outputs.append('')

    # Add the grid to the flexbox template
    template = Template('''
    <div style="display:inline-grid;grid-gap:10px 10px;grid-auto-flow: {{ 'row' if byrow else 'column' }};grid-template-rows: repeat({{nrows}}, min-content);grid-template-columns: repeat({{ncols}}, max-content);">
        {% for item in html_outputs %}
            <div style="flex: 1 1 auto;">
                {{ item }}
            </div>
        {% endfor %}
    </div>
    ''')

    # Return the html output
    return HTML(template.render(html_outputs=html_outputs, nrows=nrows, ncols=ncols, byrow=byrow))


def clean_html(notebook_as_html):
    soup = BeautifulSoup(notebook_as_html, 'html.parser')

    # removes code input blocks
    for inp in soup.find_all('div', {'class': 'input'}):
        inp.decompose()

    # removes line number container
    for inp in soup.find_all('div', {'class': 'prompt'}):
        inp.decompose()

    # removes line number output
    for inp in soup.find_all('div', {'class': 'prompt output_prompt'}):
        inp.decompose()

    # removes show/hide code toggle
    for inp in soup.find_all('input', {'type': 'submit'}):
        inp.decompose()

    # removes box shadow for better printing
    for inp in soup.find_all('div', {'id': 'notebook-container'}):
        inp['style'] = 'box shadow: none; -webkit-box-shadow: none;'

    # prevents scroll bars
    for inp in soup.find_all('div', {'class': 'output_html rendered_html output_subarea'}):
        inp['style'] = 'overflow-x: visible;'

    for inp in soup.find_all('div', {'class': 'output_html rendered_html output_subarea output_execute_result'}):
        inp['style'] = 'overflow-x: visible;'

    return str(soup)


def export_notebook_to_html(nb, notebook_filename_out, clean_notebook=True, template=None):
    if isinstance(nb, str):
        nb = _read_in_notebook(notebook_fp=nb)
    html_exporter = HTMLExporter(template_file=template) if template else HTMLExporter()
    body, resources = html_exporter.from_notebook_node(nb)
    out_fp = notebook_filename_out.replace(".ipynb", ".html")
    html_final = clean_html(body) if clean_notebook else body
    with open(out_fp, "w", encoding="utf8") as f:
        f.write(html_final)


def export_notebook_to_pdf(nb, notebook_filename_out):
    pdf_exporter = PDFExporter()
    pdf_data, resources = pdf_exporter.from_notebook_node(nb)
    out_fp = notebook_filename_out.replace(".ipynb", ".pdf")
    with open(out_fp, "wb") as f:
        f.write(pdf_data)


def _read_in_notebook(notebook_fp):
    with open(notebook_fp) as f:
        nb = nbformat.read(f, as_version=4)
    return nb


# def _add_notebook_title(nb):
#     parameters = nbparameterise.extract_parameters(nb)
#     for p in parameters:
#         if p.name == 'notebook_title':
#             nb.metadata.title = p.value
#             return
#     # default
#     nb.metadata.title = 'NoteBook'
# #


# def execute_notebook(notebook_path,
#                      notebook_path_out,
#                      params_dict,
#                      run_path='',
#                      timeout=600,
#                      export_html=True,
#                      remove_cell_tags=['remove_cell'],
#                      template=None):
#     nb = _read_in_notebook(notebook_path)
#     new_nb = _set_parameters(nb, params_dict)
#     ep = ExecutePreprocessor(timeout=timeout, kernel_name='python3')
#     tr = TagRemovePreprocessor()
#     tr.remove_cell_tags = remove_cell_tags
#
#     ep.preprocess(new_nb, {'metadata': {'path': run_path}})
#     new_nb, r = tr.preprocess(new_nb, {})
#     with open(notebook_path_out, mode='wt') as f:
#         nbformat.write(new_nb, f)
#     if export_html:
#         export_notebook_to_html(new_nb,
#                                 notebook_path_out,
#                                 clean_notebook=True,
#                                 template=template)
#     return new_nb
# #


def hide_toggle(for_next=False):
    import random

    this_cell = """$('div.cell.code_cell.rendered.selected')"""
    next_cell = this_cell + '.next()'

    toggle_text = 'Toggle show/hide'  # text shown on toggle link
    target_cell = this_cell  # target cell to control with toggle
    js_hide_current = ''  # bit of JS to permanently hide code in current cell (only when toggling next cell)

    if for_next:
        target_cell = next_cell
        toggle_text += ' next cell'
        js_hide_current = this_cell + '.find("div.input").hide();'

    js_f_name = 'code_toggle_{}'.format(str(random.randint(1,2**64)))

    html = """
        <script>
            function {f_name}() {{
                {cell_selector}.find('div.input').toggle();
            }}

            $( document ).ready(function(){{
                code_shown=false;
                $('div.input').hide()
            }});

            {js_hide_current}
        </script>

        <a href="javascript:{f_name}()">{toggle_text}</a>
    """.format(
        f_name=js_f_name,
        cell_selector=target_cell,
        js_hide_current=js_hide_current,
        toggle_text=toggle_text
    )

    return HTML(html)


def set_width_global(width=100):
    display(HTML(f"<style>.container {{ width:{width}% !important; }}</style>"))


def disable_auto_scroll():
    display(Javascript('''
        IPython.OutputArea.prototype._should_scroll = function(lines) {
            return false;
        }'''))


def create_download_link(df, filename, title='Download CSV file', index=True):
    csv = df.to_csv(index=index)
    b64 = base64.b64encode(csv.encode())
    payload = b64.decode()
    html = f'<a download="{filename}" href="data:text/csv;base64,{payload}" target="_blank">{title}</a>'
    return HTML(html)


GR_PAL = sns.diverging_palette(12, 150, s=80, l=60, n=256, as_cmap=False)
GR_CM = sns.diverging_palette(12, 150, s=80, l=60, as_cmap=True)


def create_heatmap(data, row_summary=None, col_summary=None,
                   title=None,
                   fontsize=10, float_format='.0f', figsize=(20, 10),
                   xlabel=None, ylabel=None):
    min_val, max_val = get_zero_centered_min_max(data)

    fig = plt.figure(figsize=figsize)

    ax1 = plt.subplot2grid(figsize, (0, 0), fig=fig, colspan=figsize[1]-1, rowspan=figsize[0]-1)
    ax1.xaxis.tick_top()
    sns.heatmap(data, ax=ax1,
                fmt=float_format, annot=True, annot_kws={'fontsize': fontsize},
                cmap=GR_CM, cbar=False, vmin=min_val, vmax=max_val)
    if not xlabel:
        ax1.set_xlabel('')
    if not ylabel:
        ax1.set_ylabel('')

    if title is not None:
        ax1.set_title(title)

    if col_summary is not None:
        ax2 = plt.subplot2grid(figsize, (figsize[0] - 1, 0), fig=fig, colspan=figsize[1] - 1, rowspan=1)
        if isinstance(col_summary, str):
            if col_summary.lower()=='sum':
                col_stats = pd.DataFrame(data.sum(axis=0)).T
            elif col_summary.lower()=='mean':
                col_stats = pd.DataFrame(data.mean(axis=0)).T
            else:
                raise Exception(f"col_summary={col_summary} is not supported in [sum|mean]")
        else:
            col_stats = pd.DataFrame(col_summary(data)).T

        min_val, max_val = get_zero_centered_min_max(col_stats)
        sns.heatmap(col_stats, ax=ax2,
                    fmt=float_format, annot=True, annot_kws={'fontsize': fontsize},
                    cmap=GR_CM, cbar=False, vmin=min_val, vmax=max_val, xticklabels=False, yticklabels=False)
        if not xlabel:
            ax2.set_xlabel('')
        if not ylabel:
            ax2.set_ylabel('')

    if row_summary is not None:
        ax3 = plt.subplot2grid(figsize, (0, figsize[1] - 1), fig=fig, colspan=1, rowspan=figsize[0] - 1)
        if isinstance(row_summary, str):
            if row_summary.lower()=='sum':
                row_stats = pd.DataFrame(data.sum(axis=1))
            elif row_summary.lower()=='mean':
                row_stats = pd.DataFrame(data.mean(axis=1))
            else:
                raise Exception(f"row_summary={row_summary} is not supported in [sum|mean]")
        else:
            row_stats = pd.DataFrame(row_summary(data))

        min_val, max_val = get_zero_centered_min_max(row_stats)
        sns.heatmap(row_stats.values, ax=ax3,
                    fmt=float_format, annot=True, annot_kws={'fontsize': fontsize},
                    cmap=GR_CM, cbar=False, vmin=min_val, vmax=max_val, xticklabels=False, yticklabels=False)
        if not xlabel:
            ax3.set_xlabel('')
        if not ylabel:
            ax3.set_ylabel('')

    return ax1



import pytz
import traceback
from io import StringIO
import pandas as pd
from datetime import datetime, timedelta, timezone
from qtc.consts.enums import DateDataType
from qtc.ext.logging import set_logger
logger = set_logger()


DEFAULT_TIME_ZONE = 'America/New_York'


def datetime_to_dateid(date):
    """
    :param date: Date in datetime type.
    :type date: datetime
    :return: int - Date in '%Y%m%d' format.

    >>> from datetime import datetime
    >>> import qtc.utils.datetime_utils as dtu
    >>> dtu.datetime_to_dateid(datetime(2019, 9, 25))
    20190925
    >>> adt_utc = datetime(2020, 8, 7, 2, 0, 0, tzinfo=pytz.UTC)
    >>> adt_ny = adt_utc.astimezone(pytz.timezone('America/New_York'))
    >>> dtu.datetime_to_dateid(adt_utc)
    20200807
    >>> dtu.datetime_to_dateid(adt_ny)
    20200806
    """
    formatted_date = date.strftime('%Y%m%d %Z')[:8]
    return int(formatted_date)


def dateid_to_datetime(dateid, timezone=None):
    """
    :param dateid: Date in '%Y%m%d' format.
    :type dateid: int
    :return: datetime - Date in datetime type.

    >>> import qtc.utils.datetime_utils as dtu
    >>> dtu.dateid_to_datetime(20190925)
    datetime.datetime(2019, 9, 25, 0, 0)
    """
    dt = datetime.strptime(str(dateid), '%Y%m%d')
    if timezone is not None:
        # dt = pytz.timezone(timezone).localize(dt)
        dt = dt.replace(tzinfo=pytz.timezone(timezone))

    return dt


def dateid_to_datestr(dateid, sep='-'):
    """
    :param dateid: Date in '%Y%m%d' format.
    :type dateid: int
    :param sep: Separator in the returned date str.
    :type date: int
    :return: str - Date in f"%Y{sep}%m{sep}%d" format.

    >>> import qtc.utils.datetime_utils as dtu
    >>> dtu.dateid_to_datestr(20190925)
    '2019-09-25'
    >>> dtu.dateid_to_datestr(20191013, sep='/')
    '2019/10/13'
    """
    dateid = str(dateid)
    date_str = f"{dateid[0:4]}{sep}{dateid[4:6]}{sep}{dateid[6:8]}"
    return date_str


def curr_dateid(timezone=None):
    """
    :return: int - TODAY in '%Y%m%d' format.

    >>> import qtc.utils.datetime_utils as dtu
    >>> dtu.curr_dateid()
    """
    today = datetime.now() if timezone is None else datetime.now(pytz.timezone(timezone))
    return datetime_to_dateid(today)


def is_weekday(date):
    """
    >>> import qtc.utils.datetime_utils as dtu
    >>> dtu.is_weekday(20200511)
    True
    >>> dtu.is_weekday(20200516)
    False
    """

    if isinstance(date, int):
        date = dateid_to_datetime(date)
    elif not isinstance(date, datetime):
        raise Exception(f"Supported date types are [dateid | datetime] !")

    return date.weekday() not in [5, 6]


def get_biz_dateids(start_date, end_date):
    bdates = pd.bdate_range(dateid_to_datestr(dateid=normalize_date_to_dateid(date=start_date), sep='-'),
                            dateid_to_datestr(dateid=normalize_date_to_dateid(date=end_date), sep='-')).strftime('%Y%m%d')
    return [int(date) for date in bdates]


def prev_biz_dateid(dateid):
    one_day = timedelta(days=1)
    prev_day = dateid_to_datetime(dateid) - one_day
    while not is_weekday(prev_day):
        prev_day -= one_day

    return datetime_to_dateid(prev_day)


def next_biz_dateid(dateid):
    one_day = timedelta(days=1)
    next_day = dateid_to_datetime(dateid) + one_day
    while not is_weekday(next_day):
        next_day += one_day

    return datetime_to_dateid(next_day)


def parse_datetime(dt, format=None, timezone=None, ret_as_timestamp=False):
    """
    >>> import qtc.utils.datetime_utils as dtu
    >>> dtu.parse_datetime(20200210)
    datetime.datetime(2020, 2, 10, 0, 0)
    >>> dtu.parse_datetime('2020-02-10', ret_timestamp=True)
    Timestamp('2020-02-10 00:00:00')
    >>> dtu.parse_datetime('20200210-164523', timezone='US/Eastern')
    datetime.datetime(2020, 2, 10, 16, 45, 23, tzinfo=<DstTzInfo 'US/Eastern' EST-1 day, 19:00:00 STD>)
    >>> dtu.parse_datetime('2020/02/10 16:45:23', format='%Y/%m/%d %H:%M:%S', timezone='Asia/Shanghai')
    datetime.datetime(2020, 2, 10, 16, 45, 23, tzinfo=<DstTzInfo 'Asia/Shanghai' CST+8:00:00 STD>)
    >>> dtu.parse_datetime('2020-08-06', timezone=pytz.timezone('Europe/London'))
    datetime.datetime(2020, 8, 6, 0, 0, tzinfo=<DstTzInfo 'Europe/London' BST+1:00:00 DST>)
    """

    if isinstance(dt, int):
        dt = str(dt)

    if (isinstance(dt, datetime) and not ret_as_timestamp) or \
            (isinstance(dt, pd.Timestamp) and ret_as_timestamp):
        return dt

    try:
        dt = pd.Timestamp(dt)
        if timezone is not None:
            dt = dt.tz_localize(timezone)

    except Exception as e:
        pass

    if isinstance(dt, pd.Timestamp):
        return dt if ret_as_timestamp else dt.to_pydatetime()

    if isinstance(dt, str):
        dt = pd.to_datetime(dt, format=format)
        if timezone is not None:
            dt = dt.tz_localize(timezone)

        return dt if ret_as_timestamp else dt.to_pydatetime()

    raise Exception(f"'dt' has to be of type [dateid | str | datetime | pd.Timestamp | np.datetime64] !")


def normalize_date_to_dateid(date=None):
    """
    >>> import qtc.utils.datetime_utils as dtu
    >>> dtu.normalize_date_to_dateid(date=20210929)
    20210929
    >>> dtu.normalize_date_to_dateid(date='2021-09-29')
    20210929
    >>> dtu.normalize_date_to_dateid(datetime(2021,9,29))
    20210929
    """

    if date is None:
        return curr_dateid()

    return datetime_to_dateid(parse_datetime(date))


def convert_data_type_for_date_col(df, date_col=None,
                                   from_data_type=DateDataType.DATEID, to_data_type=DateDataType.TIMESTAMP,
                                   to_date_col_idx=None):
    """
    >>> import qtc.utils.datetime_utils as dtu
    >>> dtu.convert_data_type_for_date_col(pd.DataFrame({'DateId':[20210927,20210928,20210929]}), date_col='DateId')
         DateId       Date
    0  20210927 2021-09-27
    1  20210928 2021-09-28
    2  20210929 2021-09-29
    """

    if isinstance(from_data_type, str):
        from_data_type = DateDataType.retrieve(from_data_type)

    if not isinstance(from_data_type, DateDataType):
        raise Exception(f"type(from_data_type)={type(from_data_type)} is not supported in [str|DateDataType] !")

    if isinstance(to_data_type, str):
        to_data_type = DateDataType.retrieve(to_data_type)

    if not isinstance(to_data_type, DateDataType):
        raise Exception(f"type(to_data_type)={type(to_data_type)} is not supported in [str|DateDataType] !")

    if from_data_type==to_data_type:
        # logger.warn(f'Skipping since from_data_type={from_data_type} is the same as to_data_type={to_data_type} !')
        return df

    if date_col is None:
        date_col = from_data_type.value

    if to_date_col_idx is None:
        to_date_col_idx = list(df.columns).index(date_col)+1

    from_date_col_keyword = from_data_type.value
    if 'DateTime' in date_col and (from_data_type==DateDataType.TIMESTAMP or from_data_type==DateDataType.DATETIME):
        from_date_col_keyword = 'DateTime'
    to_date_col = date_col.replace(from_date_col_keyword, to_data_type.value)

    if to_date_col in df.columns:
        df = df.drop(columns=[to_date_col])
        logger.warn(f'Column to_date_col={to_date_col} found in df and dropped !')
        to_date_col_idx -= 1

    if to_data_type==DateDataType.TIMESTAMP:
        if from_data_type==DateDataType.DATEID:
            new_dates = pd.to_datetime(df[date_col], format='%Y%m%d')
        else:
            new_dates = pd.to_datetime(df[date_col])
    elif to_data_type==DateDataType.DATEID:
        if from_data_type==DateDataType.TIMESTAMP:
            new_dates = [int(date) for date in df[date_col].dt.strftime('%Y%m%d')]
        else:
            raise Exception(f'{from_data_type} -> {to_data_type} not implemented yet!')
    else:
        raise Exception(f'{from_data_type} -> {to_data_type} not implemented yet!')

    df = df.copy()
    df.insert(to_date_col_idx, to_date_col, new_dates)
    return df


def normalize_dt(dt=None, timezone=None):
    if timezone is not None:
        if isinstance(timezone, str):
            timezone = pytz.timezone(timezone)

    if dt is None:
        dt = datetime.now() if timezone is None else datetime.now(tz=timezone)

    try:
        if isinstance(dt, str) or isinstance(dt, int):
            dt = parse_datetime(dt=dt, ret_as_timestamp=False, timezone=timezone)
        elif isinstance(dt, pd.Timestamp):
            if timezone is not None:
                dt = dt.tz_localize(timezone)
        elif isinstance(dt, datetime):
            if timezone is not None:
                dt = dt.astimezone(timezone)
    except Exception as e:
        logger.error(f"dt={dt} cannot be recognized as a datetime !")
        traceback.print_exc()
        return None

    return dt


def get_month_end(date):
    date = parse_datetime(dt=date)
    return date.replace(day=31) if date.month==12 else \
           (date.replace(month=date.month+1, day=1)-timedelta(days=1))


def get_week_start_dateid(date):
    """
    >>> import qtc.utils.datetime_utils as dtu
    >>> dtu.get_week_start_dateid(date=20220629)
    20220627
    """
    date = parse_datetime(dt=date)
    week_start_date = date - timedelta(days=date.weekday())
    return normalize_date_to_dateid(date=week_start_date)


def get_month_start_dateid(date):
    """
    >>> import qtc.utils.datetime_utils as dtu
    >>> dtu.get_month_start_dateid(date=20220629)
    20220601
    """
    date = parse_datetime(dt=date)
    return normalize_date_to_dateid(date=date.replace(day=1))


def get_year_start_dateid(date=None):
    """
    >>> import qtc.utils.datetime_utils as dtu
    >>> dtu.get_year_start_dateid(date=20220629)
    20220101
    """
    if date is None:
        date = curr_dateid()

    date = parse_datetime(dt=date)
    return normalize_date_to_dateid(date=date.replace(month=1, day=1))


def get_ttm_start_dateid(date):
    """
    >>> import qtc.utils.datetime_utils as dtu
    >>> dtu.get_ttm_start_dateid(date=20220629)
    20210629
    >>> dtu.get_ttm_start_dateid(date=20200229)
    20190228
    """
    date = parse_datetime(dt=date)
    try:
        ttm_start_date = date.replace(year=date.year-1)
    except:
        ttm_start_date = date.replace(year=date.year-1, day=date.day-1)

    return normalize_date_to_dateid(date=ttm_start_date)


def get_quarter_start_dateid(date):
    """
    >>> import sysequity_common.utils.datetime_utils as dtu
    >>> dtu.get_quarter_start_dateid(date=20220630)
    20220401
    >>> dtu.get_quarter_start_dateid(date=20220715)
    20220701
    """
    date = parse_datetime(dt=date)
    quarter_start_date = datetime(date.year, 3 * ((date.month - 1) // 3) + 1, 1)
    return normalize_date_to_dateid(date=quarter_start_date)


EXCHANGE_SESSIONS_CFG = None
def get_exchange_sessions_cfg():
    global EXCHANGE_SESSIONS_CFG
    if EXCHANGE_SESSIONS_CFG is None:
        exchange_sessions_cfg = f'''MIC,Region,Timezone,OpenTimeLocal,CloseTimeLocal
XASX,AP,Australia/Sydney,10:00:00,16:00:00
XSES,AP,Asia/Singapore,09:00:00,17:00:00
XHEL,EU,Europe/Helsinki,10:00:00,18:25:00
XLIS,EU,Europe/Lisbon,08:00:00,16:30:00
XNYS,US,America/New_York,09:30:00,16:00:00
        '''

        EXCHANGE_SESSIONS_CFG = pd.read_csv(StringIO(exchange_sessions_cfg), sep=',')

    return EXCHANGE_SESSIONS_CFG



def get_exchange_sessions(date):
    """
    >>> import sysequity_common.utils.datetime_utils as dtu
    >>> dtu.get_exchange_sessions(date=20220425)
        MIC Region          Timezone OpenTimeLocal CloseTimeLocal          OpenDateTimeLocal         CloseDateTimeLocal           OpenDateTimeUTC          CloseDateTimeUTC
    0  XASX     AP  Australia/Sydney      10:00:00       16:00:00  2022-04-25 10:00:00+10:00  2022-04-25 16:00:00+10:00 2022-04-25 00:00:00+00:00 2022-04-25 06:00:00+00:00
    1  XSES     AP    Asia/Singapore      09:00:00       17:00:00  2022-04-25 09:00:00+08:00  2022-04-25 17:00:00+08:00 2022-04-25 01:00:00+00:00 2022-04-25 09:00:00+00:00
    2  XHEL     EU   Europe/Helsinki      10:00:00       18:25:00  2022-04-25 10:00:00+03:00  2022-04-25 18:25:00+03:00 2022-04-25 07:00:00+00:00 2022-04-25 15:25:00+00:00
    3  XLIS     EU     Europe/Lisbon      08:00:00       16:30:00  2022-04-25 08:00:00+01:00  2022-04-25 16:30:00+01:00 2022-04-25 07:00:00+00:00 2022-04-25 15:30:00+00:00
    4  XNYS     US  America/New_York      09:30:00       16:00:00  2022-04-25 09:30:00-04:00  2022-04-25 16:00:00-04:00 2022-04-25 13:30:00+00:00 2022-04-25 20:00:00+00:00
    >>> dtu.get_exchange_sessions(date=20230422)
        MIC Region          Timezone OpenTimeLocal CloseTimeLocal          OpenDateTimeLocal         CloseDateTimeLocal           OpenDateTimeUTC          CloseDateTimeUTC
    0  XASX     AP  Australia/Sydney      10:00:00       16:00:00  2023-04-24 10:00:00+10:00  2023-04-24 16:00:00+10:00 2023-04-24 00:00:00+00:00 2023-04-24 06:00:00+00:00
    1  XSES     AP    Asia/Singapore      09:00:00       17:00:00  2023-04-24 09:00:00+08:00  2023-04-24 17:00:00+08:00 2023-04-24 01:00:00+00:00 2023-04-24 09:00:00+00:00
    2  XHEL     EU   Europe/Helsinki      10:00:00       18:25:00  2023-04-24 10:00:00+03:00  2023-04-24 18:25:00+03:00 2023-04-24 07:00:00+00:00 2023-04-24 15:25:00+00:00
    3  XLIS     EU     Europe/Lisbon      08:00:00       16:30:00  2023-04-24 08:00:00+01:00  2023-04-24 16:30:00+01:00 2023-04-24 07:00:00+00:00 2023-04-24 15:30:00+00:00
    4  XNYS     US  America/New_York      09:30:00       16:00:00  2023-04-24 09:30:00-04:00  2023-04-24 16:00:00-04:00 2023-04-24 13:30:00+00:00 2023-04-24 20:00:00+00:00
    >>> dtu.get_exchange_sessions(date=20230101)
        MIC Region          Timezone OpenTimeLocal CloseTimeLocal          OpenDateTimeLocal         CloseDateTimeLocal           OpenDateTimeUTC          CloseDateTimeUTC
    0  XASX     AP  Australia/Sydney      10:00:00       16:00:00  2024-01-02 10:00:00+11:00  2024-01-02 16:00:00+11:00 2024-01-01 23:00:00+00:00 2024-01-02 05:00:00+00:00
    1  XSES     AP    Asia/Singapore      09:00:00       17:00:00  2024-01-02 09:00:00+08:00  2024-01-02 17:00:00+08:00 2024-01-02 01:00:00+00:00 2024-01-02 09:00:00+00:00
    2  XHEL     EU   Europe/Helsinki      10:00:00       18:25:00  2024-01-02 10:00:00+02:00  2024-01-02 18:25:00+02:00 2024-01-02 08:00:00+00:00 2024-01-02 16:25:00+00:00
    3  XLIS     EU     Europe/Lisbon      08:00:00       16:30:00  2024-01-02 08:00:00+00:00  2024-01-02 16:30:00+00:00 2024-01-02 08:00:00+00:00 2024-01-02 16:30:00+00:00
    4  XNYS     US  America/New_York      09:30:00       16:00:00  2024-01-02 09:30:00-05:00  2024-01-02 16:00:00-05:00 2024-01-02 14:30:00+00:00 2024-01-02 21:00:00+00:00
    """

    exchange_sessions = get_exchange_sessions_cfg().copy()

    def infer_dt_local(row, time_col):
        dateid = normalize_date_to_dateid(date=date)
        dateid = dateid if is_weekday(date=date) and str(dateid)[-4:]!='0101' else next_biz_dateid(dateid=dateid)
        dt_str = f"{dateid} {row[time_col]}"
        return parse_datetime(dt_str, timezone=row['Timezone'])

    exchange_sessions['OpenDateTimeLocal'] = exchange_sessions.apply(lambda x : infer_dt_local(x, 'OpenTimeLocal'), axis=1)
    exchange_sessions['CloseDateTimeLocal'] = exchange_sessions.apply(lambda x : infer_dt_local(x, 'CloseTimeLocal'), axis=1)

    exchange_sessions['OpenDateTimeUTC'] = exchange_sessions['OpenDateTimeLocal'].apply(lambda x : x.astimezone(timezone.utc))
    exchange_sessions['CloseDateTimeUTC'] = exchange_sessions['CloseDateTimeLocal'].apply(lambda x: x.astimezone(timezone.utc))

    return exchange_sessions


def get_region_sessions(date):
    """
    >>> import sysequity_common.utils.datetime_utils as dtu
    >>> dtu.get_region_sessions(date=20230425)
                    StartDateTimeUTC            EndDateTimeUTC
    Region
    AP     2023-04-25 00:00:00+00:00 2023-04-25 09:00:00+00:00
    EU     2023-04-25 07:00:00+00:00 2023-04-25 15:30:00+00:00
    US     2023-04-25 13:30:00+00:00 2023-04-25 20:00:00+00:00
    WW     2023-04-25 00:00:00+00:00 2023-04-25 20:00:00+00:00
    >>> dtu.get_region_sessions(date=20230402)
                    StartDateTimeUTC            EndDateTimeUTC
    Region
    AP     2023-04-03 00:00:00+00:00 2023-04-03 09:00:00+00:00
    EU     2023-04-03 07:00:00+00:00 2023-04-03 15:30:00+00:00
    US     2023-04-03 13:30:00+00:00 2023-04-03 20:00:00+00:00
    WW     2023-04-03 00:00:00+00:00 2023-04-03 20:00:00+00:00
    >>> dtu.get_region_sessions(date=20240101)
                    StartDateTimeUTC            EndDateTimeUTC
    Region
    AP     2024-01-01 23:00:00+00:00 2024-01-02 09:00:00+00:00
    EU     2024-01-02 08:00:00+00:00 2024-01-02 16:30:00+00:00
    US     2024-01-02 14:30:00+00:00 2024-01-02 21:00:00+00:00
    WW     2024-01-01 23:00:00+00:00 2024-01-02 21:00:00+00:00
    """
    exchange_sessions = get_exchange_sessions(date=date)
    region_sessions = exchange_sessions.groupby('Region').apply(
        lambda x : pd.DataFrame.from_dict({'OpenDateTimeUTC':x['OpenDateTimeUTC'].min(),
                                           'CloseDateTimeUTC':x['CloseDateTimeUTC'].max()},
                                          orient='index')
    )
    region_sessions = region_sessions[0].unstack()
    region_sessions.loc['WW'] = [region_sessions['OpenDateTimeUTC'].min(), region_sessions['CloseDateTimeUTC'].max()]

    region_sessions.rename(columns={'OpenDateTimeUTC':'StartDateTimeUTC',
                                    'CloseDateTimeUTC':'EndDateTimeUTC'},
                           inplace=True)

    return region_sessions


def infer_region_cutoff_dts(dateid=None, now=None,
                            region='WW',
                            ext_minutes=None):
    """
    >>> import sysequity_common.utils.datetime_utils as dtu
    >>> dtu.infer_region_cutoff_dts(dateid=20240101)
    (Timestamp('2024-01-01 23:00:00+0000', tz='UTC'),
     Timestamp('2024-01-02 21:00:00+0000', tz='UTC'))
    >>> dtu.infer_region_cutoff_dts(dateid=20230422)
    (Timestamp('2023-04-24 00:00:00+0000', tz='UTC'),
     Timestamp('2023-04-24 20:00:00+0000', tz='UTC'))
    >>> dtu.infer_region_cutoff_dts(dateid=20230425)
    (Timestamp('2023-04-25 00:00:00+0000', tz='UTC'),
     Timestamp('2023-04-25 20:00:00+0000', tz='UTC'))
    >>> dtu.infer_region_cutoff_dts(now='2023-04-24 10:00:00+00:00', ext_minutes=15)
    (Timestamp('2023-04-23 23:45:00+0000', tz='UTC'),
     Timestamp('2023-04-24 20:15:00+0000', tz='UTC'))
    >>> dtu.infer_region_cutoff_dts(now='2023-04-24 21:00:00+00:00')
    (Timestamp('2023-04-25 00:00:00+0000', tz='UTC'),
     Timestamp('2023-04-25 20:00:00+0000', tz='UTC'))
    >>> dtu.infer_region_cutoff_dts(now='2023-04-25 01:00:00+00:00')
    (Timestamp('2023-04-25 00:00:00+0000', tz='UTC'),
     Timestamp('2023-04-25 20:00:00+0000', tz='UTC'))
    >>> dtu.infer_region_cutoff_dts(now='2024-01-01 22:00:00+00:00')
    (Timestamp('2024-01-01 23:00:00+0000', tz='UTC'),
     Timestamp('2024-01-02 21:00:00+0000', tz='UTC'))
    >>> dtu.infer_region_cutoff_dts(now='2024-01-01 23:30:00+00:00')
    (Timestamp('2024-01-01 23:00:00+0000', tz='UTC'),
     Timestamp('2024-01-02 21:00:00+0000', tz='UTC'))
    >>> dtu.infer_region_cutoff_dts(now='2024-01-02 10:00:00+00:00')
    (Timestamp('2024-01-01 23:00:00+0000', tz='UTC'),
     Timestamp('2024-01-02 21:00:00+0000', tz='UTC'))
    >>> dtu.infer_region_cutoff_dts(now='2024-01-02 22:00:00+00:00')
    (Timestamp('2024-01-02 23:00:00+0000', tz='UTC'),
     Timestamp('2024-01-03 21:00:00+0000', tz='UTC'))
    >>> dtu.infer_region_cutoff_dts(now='2024-01-03 10:00:00+00:00')
    (Timestamp('2024-01-02 23:00:00+0000', tz='UTC'),
     Timestamp('2024-01-03 21:00:00+0000', tz='UTC'))
    """
    if dateid is None:
        if now is None:
            now = datetime.now(tz=pytz.timezone('UTC'))
        if not isinstance(now, datetime):
            now = datetime.strptime(now, '%Y-%m-%d %H:%M:%S%z')
        now = now.astimezone(pytz.timezone('UTC'))

        dateid = normalize_date_to_dateid(date=now)
        start_dt, end_dt = infer_region_cutoff_dts(dateid=dateid,
                                                   region=region,
                                                   ext_minutes=ext_minutes)
        # print(f'now={now}, end_dt={end_dt}, dateid={dateid}, next_biz_dateid={next_biz_dateid(dateid=dateid)}')
        if now>end_dt:
            start_dt, end_dt = infer_region_cutoff_dts(dateid=next_biz_dateid(dateid=dateid),
                                                       region=region,
                                                       ext_minutes=ext_minutes)
    else:
        region_sessions = get_region_sessions(date=dateid)

        tz_name = 'UTC'
        if ext_minutes is None:
            ext_minutes = 0

        if region is None:
            region = 'WW'
        start_dt = region_sessions.loc[region, 'StartDateTimeUTC'].to_pydatetime() - timedelta(minutes=ext_minutes)
        end_dt = region_sessions.loc[region, 'EndDateTimeUTC'].to_pydatetime() + timedelta(minutes=ext_minutes)

        ## Backward compatibility
        try:
            end_dt = pytz.timezone(tz_name).localize(end_dt)
            start_dt = pytz.timezone(tz_name).localize(start_dt)
        except:
            pass

    start_dt = pd.Timestamp(start_dt)
    end_dt = pd.Timestamp(end_dt)
    return start_dt, end_dt


# TRADING_SESSIONS = None
# def get_trading_sessions():
#     global TRADING_SESSIONS
#     if TRADING_SESSIONS is None:
#
#         # REGION_MICS = pd.read_csv(StringIO(f'Region,FirstMIC,LastMIC\nAP,XASX,XSES\nEU,XHEL,XLIS\nUS,XNYS,XNYS'),
#         #                           index_col='Region')
#         cfg = f'''
# Region,FirstExMIC,FirstExTZ,FirstExOpenTimeLocal,FirstExCloseTimeLocal,LastExMIC,LastExTZ,LastExOpenTimeLocal,LastExCloseTimeLocal
# AP,XASX,Australia/Sydney,10:00:00,16:00:00,XSES,Asia/Singapore,09:00:00,17:00:00
# EU,XHEL,Europe/Helsinki,10:00:00,18:25:00,XLIS,Europe/Lisbon,08:00:00,16:30:00
# US,XNYS,America/New_York,09:30:00,16:00:00,XNYS,America/New_York,09:30:00,16:00:00
# WW,XASX,Australia/Sydney,10:00:00,16:00:00,XNYS,America/New_York,09:30:00,16:00:00
#         '''
#         TRADING_SESSIONS = pd.read_csv(StringIO(cfg), index_col='Region')
#
#     return TRADING_SESSIONS


# def get_region_start_end_dt(dateid=None, region=None,
#                             now=None,
#                             ext_minutes=0):
#     """
#     >>> import sysequity_common.utils.datetime_utils as dtu
#     >>> dtu.get_region_start_end_dt(dateid=20230417)
#     (Timestamp('2023-04-17 00:00:00+0000', tz='UTC'),
#      Timestamp('2023-04-17 20:00:00+0000', tz='UTC'))
#     >>> dtu.get_region_start_end_dt(dateid=20230416)
#     (Timestamp('2023-04-17 00:00:00+0000', tz='UTC'),
#      Timestamp('2023-04-17 20:00:00+0000', tz='UTC'))
#     >>> dtu.get_region_start_end_dt(dateid=20230103)
#     (Timestamp('2023-01-02 23:00:00+0000', tz='UTC'),
#      Timestamp('2023-01-03 21:00:00+0000', tz='UTC'))
#     >>> dtu.get_region_start_end_dt(now='2023-04-17 16:30:00-0400', ext_minutes=15)
#     (Timestamp('2023-04-17 23:45:00+0000', tz='UTC'),
#      Timestamp('2023-04-18 20:15:00+0000', tz='UTC'))
#     >>> dtu.get_region_start_end_dt(dateid=20230425)
#     (Timestamp('2023-04-25 00:00:00+0000', tz='UTC'),
#      Timestamp('2023-04-25 20:00:00+0000', tz='UTC'))
#     """
#     region_mics = get_region_mics()
#     if region is None:
#         first_mic, _ = region_mics.loc['AP']
#         _, last_mic = region_mics.loc['US']
#     else:
#         first_mic, last_mic = region_mics.loc[region]
#
#     import pandas_market_calendars as mcal
#     fmic = mcal.get_calendar(first_mic.upper())
#     lmic = mcal.get_calendar(last_mic.upper())
#
#     if dateid is None:
#         if now is None:
#             now = datetime.now(tz=pytz.timezone('America/New_York'))
#         else:
#             if not isinstance(now, datetime):
#                 now = datetime.strptime(now, '%Y-%m-%d %H:%M:%S%z')
#     else:
#         now = dateid_to_datetime(dateid=dateid, timezone='America/New_York')
#
#     start_dt = end_dt = None
#     while True:
#         dateid = normalize_date_to_dateid(date=now)
#         datestr = dateid_to_datestr(dateid=dateid, sep='-')
#
#         # try:
#         #     start_dt = fmic.schedule(start_date=datestr, end_date=datestr)['market_open'][0] - timedelta(minutes=ext_minutes)
#         #     end_dt = lmic.schedule(start_date=datestr, end_date=datestr)['market_close'][0] + timedelta(minutes=ext_minutes)
#         # except:
#         #     now = now.replace(hour=0, minute=0, second=0) + timedelta(days=1)
#         #     continue
#
#         fschedule = fmic.schedule(start_date=datestr, end_date=datestr)
#         if len(fschedule)>0:
#             start_dt = fschedule['market_open'][0] - timedelta(minutes=ext_minutes)
#
#         lschedule = lmic.schedule(start_date=datestr, end_date=datestr)
#         if len(lschedule)>0:
#             end_dt = lschedule['market_close'][0] + timedelta(minutes=ext_minutes)
#
#         if start_dt is None:
#             if end_dt is None:
#                 now = now.replace(hour=0, minute=0, second=0) + timedelta(days=1)
#                 continue
#         else:
#             if end_dt is None:
#                 end_dt = (start_dt.astimezone(pytz.timezone('America/New_York')) + timedelta(days=1)).replace(hour=16, minute=0, second=0) + timedelta(minutes=ext_minutes)
#
#         if now > end_dt:
#             now = now.replace(hour=0, minute=0, second=0) + timedelta(days=1)
#             continue
#
#         break
#
#     if start_dt is None:
#         start_dt = (end_dt.astimezone(pytz.timezone('Australia/Sydney')) - timedelta(days=1))
#         start_dt = start_dt.replace(hour=10, minute=0, second=0) - timedelta(minutes=ext_minutes)
#         start_dt = start_dt.astimezone(tz=pytz.timezone('UTC'))
#
#     return start_dt, end_dt

# def get_region_start_end_dt(dateid=None, region=None,
#                             now=None,
#                             ext_minutes=0):
#     if region is None:
#         region = 'WW'
#
#     session = get_trading_sessions().loc[region]
#
#     if now is None:
#         now = datetime.now(tz=pytz.timezone('America/New_York'))
#         end_dt = now.replace()
#
#     if dateid is None:
#         dateid = normalize_date_to_dateid(date=now)
#
#     start_dt = dateid_to_datetime(dateid=dateid, timezone=ftz).replace(hour=10, minute=0, second=0).astimezone(tz=pytz.timezone('UTC'))
#     end_dt = dateid_to_datetime(dateid=dateid, timezone=ltz).replace(hour=10, minute=0, second=0).astimezone(
#         tz=pytz.timezone('UTC'))




import os
import pandas as pd
import unittest
import typing as T
from qtc.ext.inspect import inspect_caller


class TestCase(unittest.TestCase):
    # def setUp(self):
    #     def dataframe_equals(df1, df2, msg=None):
    #         return df1.equals(df2)
    #     #
    #
    #     def index_equals(index1, index2, msg=None):
    #         return index1.equals(index2)
    #     #
    #     self.addTypeEqualityFunc(pd.DataFrame, dataframe_equals)
    #     self.addTypeEqualityFunc(pd.Index, index_equals)
    # #

    @staticmethod
    def load_benchmark(basename: str,
                       parse_dates: T.Optional[str] = None):
        caller_info = inspect_caller(skip=2, return_info=['FILENAME'])
        caller_file = os.path.realpath(caller_info['FILENAME'])

        benchmark_file = os.path.join(os.path.dirname(caller_file), 'benchmark', basename)
        benchmark = pd.read_csv(benchmark_file, sep=',', parse_dates=parse_dates)
        return benchmark


def main():
    unittest.main()



import pandas as pd
import unittest
import qtc.utils.misc_utils as mu


class TestMiscUtils(unittest.TestCase):
    def test_iterable_to_tuple(self):
        input = 'Hello'
        output = mu.iterable_to_tuple(input, raw_type='str')
        self.assertEqual(output, ('Hello',))

        input = '5'
        output = mu.iterable_to_tuple(input, raw_type='int')
        self.assertEqual(output, (5,))

        input = 'Hello,World'
        output = mu.iterable_to_tuple(input, raw_type='str')
        self.assertEqual(output, ('Hello','World'))

        input = '5,3,-1,3,5,9'
        output = mu.iterable_to_tuple(input, raw_type='int')
        self.assertEqual(output, (5, 3, -1, 9))

        input = '5,3,-1,3,5,9'
        output = mu.iterable_to_tuple(input, raw_type='int', remove_duplicates=False)
        self.assertEqual(output, (5,3,-1,3,5,9))

        input = '5, -1,   3,9, 9'
        output = mu.iterable_to_tuple(input, raw_type='int')
        self.assertEqual(output, (5, -1, 3, 9))

        input = '5, -1,   3,9, 9'
        output = mu.iterable_to_tuple(input, raw_type='int')
        self.assertEqual(output, (5,-1,3,9))

        input = pd.Series([4,-2,9,'test'])
        output = mu.iterable_to_tuple(input, raw_type='str')
        self.assertEqual(output, ('4', '-2', '9', 'test'))

        input = -4
        output = mu.iterable_to_tuple(input, raw_type='int')
        self.assertEqual(output, (-4,))

    def test_iterable_to_db_str(self):
        input = '5, 3,-1,3,5,9'
        output = mu.iterable_to_db_str(input, raw_type='int')
        self.assertEqual(output, '(5,3,-1,9)')

        input = 'Hello,World'
        output = mu.iterable_to_db_str(input, raw_type='str')
        self.assertEqual(output, "('Hello','World')")


if __name__ == '__main__':
    unittest.main()



from abc import ABC
from dataclasses import dataclass
from qtc.consts.enums import (AssetClass, GicsLevel, Region, RiskModel, SignalValueType,
                              Frequency, SlippageType, ExecutionType, MissingHandlingType,
                              CostModelType, RiskDataTimestamp, Enum)
from qtc.app_configs.bq.constants import nb_day_year
from dataclasses import dataclass, fields
from typing import _GenericAlias, get_args, get_origin, Literal # type: ignore

@dataclass
class ReportParams(ABC):
    """
    Contains common report parameters
    """
    calc_custom_quantile_pnl: bool = False  #: 
    calc_lead_lag: bool = True  #: 
    calc_sector_exp: bool = True  #: 
    calc_sector_pnl: bool = True  #: 
    calc_country_pnl: bool = True  #: 
    calc_country_exp: bool = True  #: 
    calc_factor_exp: bool = True  #: 
    calc_factor_risk: bool = True  #: 
    calc_position_quantile_pnl: bool = True  #: 
    calc_volume_quantile_pnl: bool = True  #: 
    calc_fw_vol: bool = False  #: 
    calc_adv: bool = True  #: 
    calc_beta: bool = True  #: 
    calc_long_short_pnl: bool = True  #: 
    calc_earnings_pnl: bool = False  #: 
    calc_day_of_week_pnl: bool = True  #: 
    calc_quarterly_stats: bool = True  #: 
    calc_top_drawdown: bool = True  #: 
    calc_stock_bias: bool = True  #: 
    calc_factor_bias: bool = True  #: 
    calc_top_bottom_stock: bool = True  #:
    calc_top_factor: bool = True  #:
    calc_pnl_decomposition: bool = True  #: 
    calc_turnover: bool = True  #: 
    calc_position_stats: bool = True  #: 
    calc_trades_stats: bool = True  #: 
    calc_num_active_positions: bool = True  #: 
    calc_pnl: bool = True  #: 
    calc_accounting_costs: bool = True  #: 
    calc_exposures: bool = True  #: 
    calc_vol_hist_from_daily_pnl: bool = True  #: 
    calc_vol_hist_from_daily_factor_pnl: bool = True  #:
    calc_aggregate_accounting_stats: bool = True  #: 
    apply_holiday_adjustment: bool = False  #: 
    calc_market_cap_breakdown: bool = True  #

    vol_lookback_window: int = nb_day_year  #: 
    nb_top_bottom_stock: int = 50  #: number of stocks for which we want to report performance
    custom_quantile_bins: int = 5  #: number of bins for custom quantile pnl computation
    max_lead_days: int = 2  #: lead/lag pnl computation
    max_lag_days: int = 10  #: lead/lag pnl computation
    nb_top_drawdown: int = 10  #: number of drawdowns to report (sorted by depth)
    nb_day_significance_bias: int = nb_day_year  #: Do not report bias for stocks with few valid positions
    nb_stock_bias: int = 50  #: number of stocks for which we report the bias

    drift_pos: bool = True  #: account for positions drift; drift means that appreciation or depreciation in value doesn't cause trades

    gics_level: GicsLevel = GicsLevel.Sector  #: 
    
    risk_data_timestamp: RiskDataTimestamp = RiskDataTimestamp.pos_t  #: Shift risk data

    archive_run: bool = False  #: 
    
    execution_type: ExecutionType = ExecutionType.Vwap  #: 

    calc_dollar: bool = False  #: Set to True if want stats in dollars and not as percentage of gmv


@dataclass
class SignalParams(ABC):
    """
    Contains common signal parameters
    """
    asset_class: AssetClass = AssetClass.EQY  #: 
    region: Region = Region.US  #: 
    risk_model: RiskModel = RiskModel.AXUS4_MH  #: 
    signal_value_type: SignalValueType = SignalValueType.MoneyAllocation  #: 
    frequency: Frequency = Frequency.D1  #: 
    missing_dates_handling: MissingHandlingType = MissingHandlingType.Ignore  #:
    
@dataclass
class CostParamsBase(ABC):
    slippage_type: SlippageType = SlippageType.MarketImpact  #:
    cost_model_type: CostModelType = CostModelType.Equity  #: 


def dict2dataclass(cls, dict_args: dict):
    typed_dict = {}
    for f in fields(cls):
        if not f.init:
            continue
        
        value = dict_args.get(f.name)
        if value is None:
            continue
        if get_origin(f.type):
            if isinstance(value, dict):
                typed_dict[f.name] = get_args(f.type)[0](**value)
            else:
                typed_dict[f.name] = get_args(f.type)[0](value)
        else:
            typed_dict[f.name] = f.type(value)

    return cls(**typed_dict)


def dataclass2dict(cls) -> dict:
    res = {}
    for f in fields(cls):
        if not f.init:
            continue
        if issubclass(type(f.type), _GenericAlias):
            res[f.name] = getattr(cls, f.name)
        elif issubclass(f.type, Enum):
            res[f.name] = getattr(cls, f.name).value
        else:
            res[f.name] = getattr(cls, f.name)

    return res




import qtc.ext.unittest as ut
import qtc.utils.db_utils as dbu
from qtc.consts.enums import DBType
import pandas as pd


class TestDBUtils(ut.TestCase):
    def test__infer_db_type_name(self):
        self.assertEqual(dbu._infer_db_type_name(db_code='GMBP-RDS'), (DBType.MYSQL, 'GMBP-RDS'))

    def test__process_cols(self):
        self.assertEqual(dbu._process_cols(available_cols='SecurityId,AxiomaId,BbergCode,NMV,Return',
                                           required_cols='SecurityId,NMV,BbergCode'),
                         ['SecurityId', 'NMV', 'BbergCode', 'AxiomaId', 'Return'])
        self.assertEqual(dbu._process_cols(available_cols='SecurityId,AxiomaId,BbergCode,NMV,Return',
                                           required_cols='SecurityId'),
                         ['SecurityId', 'AxiomaId', 'BbergCode', 'NMV', 'Return'])
        self.assertEqual(dbu._process_cols(available_cols='SecurityId,AxiomaId,BbergCode,NMV,Return',
                                           required_cols='SecurityId',
                                           expected_cols='NMV,Return'),
                         ['SecurityId', 'NMV', 'Return'])
        self.assertEqual(dbu._process_cols(available_cols='SecurityId,AxiomaId,BbergCode,NMV,Return',
                                           required_cols='SecurityId,NMV',
                                           expected_cols='NMV,Return,SecurityId'),
                         ['SecurityId', 'NMV', 'Return'])

        self.assertEqual(dbu._process_cols(available_cols='SecurityId,AxiomaId,BbergCode,NMV,Return',
                                           expected_cols='SecurityId,NMV'),
                         ['SecurityId', 'NMV'])

    def test_compile_sql_where_clauses(self):
        self.assertEqual(dbu.compile_sql_where_clause(), '')
        self.assertEqual(dbu.compile_sql_where_clause([]), '')
        self.assertEqual(dbu.compile_sql_where_clause("CountryCode IN ('US')"), "WHERE CountryCode IN ('US')")
        self.assertEqual(dbu.compile_sql_where_clause(['SecurityId IN (407,409)', 'FactorModelId=60']), 'WHERE SecurityId IN (407,409)\n\t\tAND FactorModelId=60')
        self.assertEqual(dbu.compile_sql_where_clause("EffectiveStartDate<='2021-09-23'||EffectiveEndDate>='2020-09-23'"), "WHERE EffectiveStartDate<='2021-09-23'\n\t\tAND EffectiveEndDate>='2020-09-23'")
        self.assertEqual(dbu.compile_sql_where_clause(['SecurityId IN (407,409)', 'FactorModelId=60', ' ']), 'WHERE SecurityId IN (407,409)\n\t\tAND FactorModelId=60')


if __name__ == '__main__':
    ut.main()



import numpy as np
# Accounting annualization factors
nb_day_year = 252  #: average trading days in US
sqrt_nb_day_year = nb_day_year**0.5  #: square root of number of trading days

mkt_cap_bins = [-np.inf, 100e6, 200e6, 300e6, 1e9, 10e9, np.inf]



import pandas as pd

# def drawdown_calc(df, num_drawdowns, date_col='Date',
#                   value_col='Return', is_date_indexed=False):
#     """This is a util method to calculate the drawdowns from a series of date-indexed returns values
#
#     :param df: dataframe containing return or gross values with dates
#     :param num_drawdowns: number of drawdowns
#     :param date_col: specify the column name for dates
#     :param value_col:  column to calc drawdowns on
#     :param is_date_indexed: True if the dataframe is indexed by date
#
#     :return: A list of [[drawdown_begin_date, drawdown_end_date, drawdown_value]]
#     """
#     returns = df[value_col][1:]
#     returns = list(returns)
#     drawdown_list = []
#     rolling_drawdown = _recur_drawdown_calc(returns, drawdown_list, 0)
#     drawdown_list.sort(key=lambda x: x[2])
#     drawdown_list = drawdown_list[:num_drawdowns]
#     # Convert raw-index to dates
#     for dd in drawdown_list:
#         if is_date_indexed:
#             dd[0] = df.index[dd[0] + 1]
#             dd[1] = df.index[dd[1]]
#         else:
#             dd[0] = df[date_col].iloc[dd[0] + 1]
#             dd[1] = df[date_col].iloc[dd[1]]
#
#     # Remove all the dds with only 1 day length.
#     drawdown_list = [x for x in drawdown_list if x[0] != x[1]]
#     # logging.info("The drawdowns are computed as follows - {}".format(drawdown_list))
#     return drawdown_list, rolling_drawdown
#
#
# def _recur_drawdown_calc(returns, drawdown_list, add_to_index=0):
#     length = len(returns)
#     if length <= 1:
#         return None
#     values = [1.0 for i in range(length + 1)]
#     rolling_max = [-1 for i in range(length + 1)]
#     rolling_dd = [0 for i in range(length + 1)]
#
#     rolling_max[0] = 1
#     rolling_high_index = 0
#     max_dd = 0
#     trough_index = 0
#     peak_index = 0
#
#     # calc maxDD
#     for i in range(length):
#         values[1 + i] = values[i] * (1.0 + returns[i])
#         if values[1 + i] > rolling_max[i]:
#             rolling_high_index = 1 + i
#         rolling_max[1 + i] = max(rolling_max[i], values[1 + i])
#         rolling_dd[1 + i] = values[1 + i] / rolling_max[1 + i] - 1.0
#
#         if rolling_dd[1 + i] < max_dd:
#             max_dd = rolling_dd[1 + i]
#             trough_index = 1 + i
#             peak_index = rolling_high_index
#     peak_index += add_to_index
#     trough_index += add_to_index
#     if peak_index == trough_index:
#         return None
#     drawdown_list.append([peak_index, trough_index, max_dd])
#
#     # recursion steps
#     _recur_drawdown_calc(returns[: peak_index - add_to_index], drawdown_list, add_to_index)
#     _recur_drawdown_calc(returns[trough_index - add_to_index + 1:], drawdown_list, trough_index + 1)
#
#     return rolling_dd


def calc_cum_nmvs(returns):
    returns = list(returns).copy()
    returns.insert(0, 0.0)
    cum_nmvs = (pd.Series(returns) + 1).cumprod()

    return cum_nmvs


def _calc_drawdowns(returns, num_drawdowns=None):
    # get cumulative NMVs
    cum_nmvs = calc_cum_nmvs(returns)

    N = len(returns)

    # calc all candidates with cum_rets(i->j)<0
    drawdown_candidates = list()
    for i in range(N):
        candidates = list()
        for j in range(i, N):
            cum_return = cum_nmvs[j] / cum_nmvs[i] - 1
            if cum_return < 0:
                candidates.append([i, j, cum_return])

        drawdown_candidates.extend(candidates)

    if len(drawdown_candidates)==0:
        return drawdown_candidates

    # sort all candidates by cum_rets
    drawdown_candidates.sort(key=lambda x: x[2], reverse=False)

    # remove overlaps
    #     drawdowns = [drawdown_candidates[0]]
    #     N = len(drawdown_candidates)
    drawdowns = list()
    if num_drawdowns is None:
        num_drawdowns = 1e10

    for candidate in drawdown_candidates:  # for each candidate
        if not drawdowns:
            # add the first candiate anyway
            drawdowns.append(candidate)
            continue

        overlap = False
        # check exisiting drawdowns
        for drawdown in drawdowns:
            # if curr candidate has NO overlap with this existing drawdown, the continue
            if candidate[1] < drawdown[0] or candidate[0] > drawdown[1]:
                continue
            else:  # otherwise, overlap found, break here!
                overlap = True
                break

        # No overlaps with all existing drawdowns, then add this candidate to drawdown list
        if not overlap:
            drawdowns.append(candidate)

        if len(drawdowns)>=num_drawdowns:
            break

    return drawdowns


def calc_drawdowns(df, date_col='Date', return_col='Return',
                   num_drawdowns=None,
                   ret_as_df=True):
    if date_col=='index':
        df.sort_index(inplace=True)
        dates = df.index.values
    else:
        df.sort_values(date_col, inplace=True)
        dates = df[date_col].values

    drawdowns = _calc_drawdowns(df[return_col], num_drawdowns=num_drawdowns)

    for i, [start, end, dd] in enumerate(drawdowns):
        drawdowns[i][0] = pd.Timestamp(dates[start])
        drawdowns[i][1] = pd.Timestamp(dates[end])

    if ret_as_df:
        drawdowns = pd.DataFrame(drawdowns, columns=['StartDate', 'EndDate', 'DD'])

    return drawdowns



import os
import shutil
import pandas as pd
import qtc.ext.unittest as ut
import qtc.file.file_manager as fileman


class CustomFileManager(fileman.FileManager):
    NAME = 'CUSTOM_FILE_MANAGER'
    DATETIME_FOLDERS_BUILDER = fileman.DateTimeFoldersBuilderManager.get(
        root=os.path.join(os.path.dirname(__file__), 'test_custom_file_manager'),
        datetime_folder_structure_mode='DATEID',
        timezone='America/New_York')
    FILE_SERIALIZER_DICT = {'POSITIONS': fileman.FileSerializerManager.get(filename_prefix='positions',
                                                                           filename_dt_component_type='DATEID',
                                                                           protocol='PKL.GZ',
                                                                           filename_component_sep='.')}
CustomFileManager.register()


class TestFileManager(ut.TestCase):
    def test_get_meta(self):
        benchmark = self.load_benchmark(basename='file_manager_meta.csv')
        benchmark.set_index(['EnumClass','EnumName'], inplace=True)
        pd.testing.assert_frame_equal(fileman.get_meta(), benchmark)

    def test_file_manager_serialization(self):
        cfm = fileman.FileManagerRegistry.get('CUSTOM_FILE_MANAGER')

        data = pd.DataFrame({'SecurityId': [409, 2654], 'NMV': [-10, 10]})
        file_path = cfm.serialize(data=data,
                                  file_serializer='POSITIONS',
                                  dt=20210929,
                                  subfolders=None,
                                  extra_filename_components=None,
                                  include_run_dt=False)
        data = pd.DataFrame({'SecurityId': [409,2654], 'NMV': [-20,20]})
        file_path = cfm.serialize(data=data,
                                  file_serializer='POSITIONS',
                                  dt=20210930,
                                  subfolders=None,
                                  extra_filename_components=None,
                                  include_run_dt=False)

        target = cfm.deserialize(file_serializer='POSITIONS',
                                 start_dt=20210901, end_dt=20210930,
                                 subfolders=None,
                                 extra_filename_components=None,
                                 allow_as_of=False,
                                 dt_type='DateId')

        benchmark = self.load_benchmark(basename='deserialization.param1.csv')
        pd.testing.assert_frame_equal(target, benchmark)

        target = cfm.deserialize(file_serializer='POSITIONS',
                                 dts='20210929,20210930',
                                 subfolders=None,
                                 extra_filename_components=None,
                                 allow_as_of=False,
                                 dt_type='DateId')
        pd.testing.assert_frame_equal(target, benchmark)

        # test_data_root = sysequity_dir('test_custom_file_manager')
        test_data_root = cfm.root#'test_custom_file_manager'
        shutil.rmtree(test_data_root)


if __name__ == '__main__':
    ut.main()



import os
import typing as T
from types import TracebackType #pylint: disable=no-name-in-module
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.text import MIMEText
import qtc.env_config as ecfg
import qtc.utils.misc_utils as mu
from qtc.ext.logging import set_logger
logger = set_logger()

USER_FULLNAME_MAP = {}


def normalize_email(email_address: str, domain: str) -> str:
    """
    >>> import qtc.utils.email_utils as emailu
    >>> emailu.normalize_email('ahu')
    'ahu@***.com'
    """
    at_domain = f'@{domain}'
    if email_address.lower().endswith(at_domain):
        return email_address

    if '@' in email_address:
        raise Exception(f'Invalid character "@" found in email {email_address}. '
                        f'Valid email addresses either are simple user names (ie. "jd") '
                        f'or end with {at_domain} (ie. jd{at_domain}).')

    user_full_name = USER_FULLNAME_MAP.get(email_address, None)
    email_address += at_domain
    return email_address if user_full_name is None else f'{user_full_name}<{email_address}>'


def normalize_email_list(email_list):
    if email_list is None:
        return list()

    return [normalize_email(ea) for ea in mu.iterable_to_tuple(email_list, raw_type='str') if len(ea) > 0]


def get_email_meta_field(field: str,
                         email_app=None,
                         default: str = None,
                         ) -> str:
    """
    >>> import qtc.utils.email_utils as emailu
    >>> emailu.get_email_meta_field(field='from', email_app='hedge_optimizer')
    'QTHO'
    >>> emailu.get_email_meta_field(field='from')
    """
    cfg = ecfg.get_env_config()
    default = cfg.get(f'email.{field}', default)
    if email_app is None:
        return default
    #
    # email_app_name = email_app.value if isinstance(email_app, ecfg.EmailApp) else email_app
    email_app_name = email_app
    return cfg.get(f'email.{email_app_name}.{field}', default)


def format_email_subject(subject: str, email_app=None) -> str:
    """
    Formats an email subject for a given EmailApp as specified for this environment.

    In most cases, that means emails in DEV will be:

    '[DEV] email subject'

    ... emails in UAT will be:

    '[UAT] email subject'

    ... and emails in PROD will be:

    'email subject'
    >>> import qtc.utils.email_utils as emailu
    >>> emailu.format_email_subject(subject='test subject', email_app='hedge_optimizer')
    """
    subject_fmt = get_email_meta_field(field='subject_fmt', email_app=email_app)
    if subject_fmt is None:
        return subject
    return subject_fmt.format(subject=subject)


class EmailMeta:
    def __init__(self, email_app=None,
                 smtp_server = None,
                 from_address = None,
                 to_address_list = None,
                 cc_address_list = None,
                 bcc_address_list = None):
        self.smtp_server = get_email_meta_field(field='smtp_server', email_app=email_app,
                                                default=smtp_server)
        self.from_address = normalize_email(email_address=get_email_meta_field(field='from', email_app=email_app,
                                                                               default=from_address))
        self.to_address_list = normalize_email_list(get_email_meta_field(field='to', email_app=email_app,
                                                                         default=to_address_list))
        self.cc_address_list = normalize_email_list(get_email_meta_field(field='cc', email_app=email_app,
                                                                         default=cc_address_list))
        self.bcc_address_list = normalize_email_list(get_email_meta_field(field='bcc', email_app=email_app,
                                                                          default=bcc_address_list))

    def __repr__(self):
        return f'SMTP: {self.smtp_server}\n' \
               f'From: {self.from_address}\n' \
               f'To  : {self.to_address_list}\n' \
               f'CC  : {self.cc_address_list}\n' \
               f'BCC : {self.bcc_address_list}\n'


class HtmlEmail:
    def __init__(self,
                 subject: str,
                 email_app=None,
                 *,
                 smtp_server = 'smtp', # set default smtp_server here!!!
                 from_address = None,
                 to_address_list = None,
                 cc_address_list = None,
                 bcc_address_list = None
                 ) -> None:
        # self._email_app = email_app
        self._subject = format_email_subject(subject=subject, email_app=email_app)
        self._email_meta = EmailMeta(email_app=email_app,
                                     smtp_server=smtp_server,
                                     from_address=from_address,
                                     to_address_list=to_address_list,
                                     cc_address_list=cc_address_list,
                                     bcc_address_list=bcc_address_list)

        # self._email_app_name = email_app.value if isinstance(email_app, ecfg.EmailApp) else email_app
        self._email_app_name = email_app

        msg = MIMEMultipart('alternative')
        msg['Subject'] = self._subject
        msg['From'] = self._email_meta.from_address
        msg['To'] = ','.join(self._email_meta.to_address_list)
        msg['Cc'] = ','.join(self._email_meta.cc_address_list)
        msg['Bcc'] = ','.join(self._email_meta.bcc_address_list)

        self._msg = msg

    def add_html(self, html: str) -> None:
        text = MIMEText(html, 'html')
        self._msg.attach(text)

    def attach_file(self, attachment_name: str, attachment) -> None: #type: ignore
        part = MIMEApplication(attachment.read(), Name=attachment_name)
        part['Content-Disposition'] = f'attachment; filename="{attachment_name}"'
        self._msg.attach(part)

    def attach_file_from_filesystem(self,
                                    filename: str,
                                    *,
                                    attachment_name = None
                                    ) -> None:
        with open(filename, mode="rb") as attachement:
            if attachment_name is None:
                attachment_name = os.path.basename(filename)
            self.attach_file(attachment_name, attachement)

    def __enter__(self): #type: ignore
        return self

    def __exit__(self, exc_type: type, exc_val: T.Any, exc_tb: TracebackType) -> None:
        title = f'Email: {self._subject}'
        if self._email_app_name is not None:
            title += f' ({self._email_app_name})'

        if exc_val is None:
            logger.info(f"Email SMTP: ({self._email_meta.smtp_server})")
            logger.info(f"Email From: ({self._msg['From']})")
            logger.info(f"Email To  : ({self._msg['To']})")
            logger.info(f"Email Cc  : ({self._msg['Cc']})")
            logger.info(f"Email BCc : ({self._msg['Bcc']})")

            smtp = smtplib.SMTP(self._email_meta.smtp_server)
            smtp.send_message(self._msg)
            logger.info(f'Sent {title}')
            smtp.quit()
        else:
            logger.error(f'Failed in sending {title} with exc_val={exc_val}, exc_tb={exc_tb} !')


import os, inspect
import pandas as pd
from qtc.ext.configurable import Configurable
import qtc.file.file_manager as fileman
import qtc.utils.datetime_utils as dtu
from qtc.consts.enums import FileCacheNameFormat, FileCacheMode
from qtc.ext.logging import set_logger
logger = set_logger()


class FileCache(Configurable, fileman.FileManager):
    DEFAULT_CONFIG = {'file_cache_mode': FileCacheMode.DISABLED,
                      'file_cache_dir': os.getenv('FILE_CACHE_ROOT', default=None),
                      'file_cache_name': FileCacheNameFormat.FUNC_NAME_KWARGS,
                      'file_cache_protocol': 'CSV'}

    NAME = 'FILE_CACHE_FILE_MANAGER'
    DATETIME_FOLDERS_BUILDER = fileman.DateTimeFoldersBuilderManager.get(root='',
                                                                         datetime_folder_structure_mode='NON_DATETIME')

    def collect_defaults(self):
        self.defaults = dict()
        if self.func is None:
            return

        signature = inspect.signature(self.func)
        for name, p in signature.parameters.items():
            if p.default!=inspect._empty:
                self.defaults[name] = p.default

    def __init__(self, func=None, datetime_folders_builder=None, file_serializer_dict=None):
        Configurable.__init__(self)
        fileman.FileManager.__init__(self,
                                     datetime_folders_builder=datetime_folders_builder,
                                     file_serializer_dict=file_serializer_dict)
        self.func = func

        self.collect_defaults()

    def compile_file_cache_name(self, **kwargs):
        file_cache_name = self.config['file_cache_name']
        if isinstance(file_cache_name, str):
            try:
                file_cache_name = FileCacheNameFormat.retrieve(file_cache_name)
            except:
                pass

        if isinstance(file_cache_name, str):
            for k, v in kwargs.items():
                if v is None:
                    continue
                    
                if k in ['dateid','date']:
                    dateid = dtu.normalize_date_to_dateid(date=v)
                    for field in ['dateid', 'date']:
                        ks = '${' + field + '}'
                        file_cache_name = file_cache_name.replace(ks, str(dateid))
                else:
                    ks = '${' + k + '}'
                    file_cache_name = file_cache_name.replace(ks, str(v))

            return file_cache_name

        if file_cache_name==FileCacheNameFormat.FUNC_NAME:
            return self.func.__name__
        elif file_cache_name==FileCacheNameFormat.FUNC_NAME_DATEID:
            file_cache_name = self.func.__name__
            if 'dateid' in kwargs:
                file_cache_name += '.' + str(kwargs['dateid'])
            elif 'date' in kwargs:
                dateid = dtu.normalize_date_to_dateid(date=kwargs['date'])
                file_cache_name += '.' + str(dateid)

            return file_cache_name
        elif file_cache_name==FileCacheNameFormat.FUNC_NAME_KWARGS:
            file_cache_name = self.func.__name__
            for k in kwargs:
                file_cache_name += '.' + f'{k}={kwargs[k]}'

            return file_cache_name
        # if file_cache_name==FileCacheNameFormat.FUNC_NAME_KWARGS_DATEID:
        #     file_cache_name = self.func.__name__
        #     for k, v in kwargs.items():
        #         file_cache_name += '.' + f'{k}={v}'
        #     #
        #     if 'dateid' in kwargs:
        #         file_cache_name += '.' + kwargs['dateid']
        #     #
        #     return file_cache_name
        #
        else:
            raise

    def __call__(self, *args, **kwargs):
        self.config.update(self.defaults)
        keys_config, keys_non_config = self.update_config(**kwargs)
        keys_non_config = {k: self.config[k] for k in self.config if k not in FileCache.DEFAULT_CONFIG}

        file_cache_mode = self.config['file_cache_mode']
        if isinstance(file_cache_mode, str):
            file_cache_mode = FileCacheMode.retrieve(file_cache_mode)

        if file_cache_mode==FileCacheMode.DISABLED:
            logger.debug(f'Skipping file caching {self.func.__module__}.{self.func.__name__} since file_cache_mode={file_cache_mode} !')
            self.reset_config()
            return self.func(*args, **kwargs)

        root = self.config['file_cache_dir']
        if root is None:
            root = os.getcwd()

        file_manager = fileman.FileManagerRegistry.get(self.NAME, root=root)

        # file_cache_name = self.compile_file_cache_name(**{k: v for k,v in keys_non_config.items() if v is not None})
        file_cache_name = self.compile_file_cache_name(**{k: v for k, v in kwargs.items() if k in keys_non_config and v is not None})
        file_serializer = fileman.FileSerializerManager.get(filename_prefix=file_cache_name,
                                                            filename_dt_component_type='NON_DATETIME',
                                                            protocol=self.config['file_cache_protocol'],
                                                            index=False)

        # file_path = file_manager.search_file(file_serializer=file_serialier,
        #                                      dt=None, subfolders=None, is_subfolders_under_dt_folder=True,
        #                                      extra_filename_components=None,
        #                                      allow_as_of=False,
        #                                      log_error=True)
        file_path = file_manager.compile_file_path(file_serializer=file_serializer,
                                                   dt=None, subfolders=None, is_subfolders_under_dt_folder=True,
                                                   extra_filename_components=None,
                                                   include_run_dt=True)

        if file_cache_mode==FileCacheMode.OVERWRITE:
            data = self.func(*args, **kwargs)
            file_path = file_manager.serialize(data=data,
                                               file_serializer=file_serializer,
                                               dt=None, subfolders=None, is_subfolders_under_dt_folder=True,
                                               extra_filename_components=None,
                                               include_run_dt=False)
            self.reset_config()
            return data

        if file_cache_mode == FileCacheMode.ENABLED:
            if not os.path.exists(file_path):
                logger.debug(f'FileCache does not exist ! Loading data ...')
                data = self.func(*args, **kwargs)
                file_path = file_manager.serialize(data=data,
                                                   file_serializer=file_serializer,
                                                   dt=None, subfolders=None, is_subfolders_under_dt_folder=True,
                                                   extra_filename_components=None,
                                                   include_run_dt=False)
            else:
                data = file_manager._deserialize(file_serializer=file_serializer,
                                                 dt=None, subfolders=None, is_subfolders_under_dt_folder=True,
                                                 extra_filename_components=None,
                                                 allow_as_of=False,
                                                 dt_col=None,
                                                 log_error=True)

            self.reset_config()
            return data

        if file_cache_mode==FileCacheMode.READONLY:
            if not os.path.exists(file_path):
                self.reset_config()
                raise Exception(f'FileCache {file_path} does not exist while file_cache_mode={file_cache_mode} !')

            data = file_manager._deserialize(file_serializer=file_serializer,
                                             dt=None, subfolders=None, is_subfolders_under_dt_folder=True,
                                             extra_filename_components=None,
                                             allow_as_of=False,
                                             dt_col=None,
                                             log_error=True)
            self.reset_config()
            return data

FileCache.register()


def get_meta():
    file_cache_mode_enums = FileCacheMode.get_meta()
    file_cache_mode_enums['EnumClass'] = 'FileCacheMode'
    file_cache_name_format_enums = FileCacheNameFormat.get_meta()
    file_cache_name_format_enums['EnumClass'] = 'FileCacheNameFormat'
    protocol_enums = fileman.FileSerializationProtocol.get_meta()
    protocol_enums = protocol_enums.loc[['CSV','PKL','PKL_GZ']]
    protocol_enums['EnumClass'] = 'FileSerializationProtocol'

    meta = pd.concat([file_cache_mode_enums,
                      file_cache_name_format_enums,
                      protocol_enums])
    meta.index.name = 'EnumName'
    return meta.reset_index().set_index(['EnumClass','EnumName'])


import yaml
from qtc.ext.logging import set_logger
logger = set_logger()


class Configurable:
    DEFAULT_CONFIG = dict()

    @property
    def config(self):
        return self.__config

    @config.setter
    def config(self, config):
        self.__config = config

    def update_config(self, **kwargs):
        config = self.DEFAULT_CONFIG.copy()
        config.update(self.config)

        keys_config = set()
        keys_non_config = set()
        for key, value in kwargs.items():
            if key not in config:
                logger.debug(f'keyword={key} in kwargs not recognized in default_config !')
                keys_non_config.add(key)
            else:
                keys_config.add(key)

        config.update(kwargs)
        self.config = config.copy()

        return keys_config, keys_non_config

    def reset_config(self):
        self.config = self.DEFAULT_CONFIG.copy()

    def _read_config(self, config):
        if isinstance(config, str):
            try:
                with open(config) as fh:
                    config = yaml.safe_load(fh, Loader=yaml.FullLoader)
            except Exception as e:
                logger.error(f"Failed to load yaml config from {config}:\n{e}")
                return None

        self.config = config

    def __init__(self, config=dict(), **kwargs):
        self._read_config(config=config)
        self.update_config(**kwargs)



import sys
import logging
from qtc.ext.inspect import inspect_caller


# DEFAULT_FORMAT = '[%(levelname).1s %(asctime)s %(module)s:%(funcName)s] - %(message)s'
DEFAULT_FORMAT = '[%(levelname).1s %(asctime)s %(filename)s:%(lineno)d] - %(message)s'


def set_logger(name: str = None,
               format: str = DEFAULT_FORMAT,
               logfile: str = None):
    """Create an easy logging interface. By default it prints logging messages to stderr (not stdout).

    :param name: Logging name with period-separated hierarchy. Default: module.__name__ .
    :param format: Logging format str.
    :param logfile: If provided, logging will be directed to it.
    :return: None

    >>> from qtc.ext.logging import set_logger
    >>> import os
    >>> logger = set_logger(logfile=os.path.join(os.environ['HOME'], 'logs', 'test_logger.log'))
    >>> logger.debug('This helps debugging')
    >>> logger.info('This is running')
    >>> logger.warning('This is a warning')
    >>> logger.error('This is an error')
    >>> logger.critical('This is a critical situation')

    .. note::
        To suppress logging messages at certain level, do something like:
        logging.getLogger('module.full.hierarchy').setLevel(logging.CRITICAL)
    """

    formatter = logging.Formatter(fmt=format)

    stream_handler = logging.StreamHandler(stream=sys.stderr)
    stream_handler.setFormatter(formatter)

    logger = logging.getLogger(name or inspect_caller(skip=2, return_info=['MODULE'])['MODULE'])
    logger.setLevel(logging.INFO)
    if not logger.handlers:
        logger.addHandler(stream_handler)

    # File logging
    if logfile:
        file_handler = logging.FileHandler(logfile, mode='a')
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    return logger


if __name__ == '__main__':
    logger = set_logger()
    logger.info('This is a test')
    print(logging.getLogger('__main__'))

    import os
    logger = set_logger(logfile=os.path.join(os.environ['HOME'], 'logs', 'test_logger.log'))
    logger.info('This is running')


import qtc.ext.unittest as ut
from qtc.ext.enum import Enum


class TestEnum(ut.TestCase):
    def test_retrieve_by_name(self):
        class RiskRegionEnum(Enum):
            AP = 'Asia'
            EU = 'EU'
            US = 'US'
            WW = 'WW'
            ROW = 'ROW'

        self.assertEqual(RiskRegionEnum.retrieve_by_name('AP'), RiskRegionEnum.AP)
        self.assertEqual(RiskRegionEnum.retrieve_by_name('EU'), RiskRegionEnum.EU)
        self.assertEqual(RiskRegionEnum.retrieve_by_name('US'), RiskRegionEnum.US)

    def test_retrieve(self):
        class Symbology(Enum):
            BARRA_ID = 'BarraId'
            AXIOMA_ID = 'AxiomaId'
            FACTSET_ID = 'FactsetId'

        self.assertEqual(Symbology.retrieve('BarraId'), Symbology.BARRA_ID)
        self.assertEqual(Symbology.retrieve('AXIOMA_ID'), Symbology.AXIOMA_ID)
        self.assertEqual(Symbology.retrieve('FactsetId'), Symbology.FACTSET_ID)


if __name__ == '__main__':
    ut.main()



from qtc.ext.enum import Enum


class DataTransmissionProtocol(Enum):
    TXT = 'txt'
    JSON = 'json'
    PICKLE = 'pickle'
    XML = 'xml'


class DollarDenominator(Enum):
    ONE = 'one'
    MILLION = 'million'


class DecimalUnit(Enum):
    DECIMAL = 'decimal'
    PERCENTAGE = '%'
    PERCENTAGE_SQUARE = '%^2'
    BPS = 'bps'


class DateDataType(Enum):
    DATEID = 'DateId'
    DATESTR = 'DateStr'
    TIMESTAMP = 'Date'
    DATETIME = 'Date'


class DataSource(Enum):
    DATABASE = 'DB'
    FMS = 'FMS'
    FILE_CACHE = 'FC'
    RERDSP = 'RERDSP'
    API = 'API'


class FileCacheNameFormat(Enum):
    FUNC_NAME = 'func_name'
    FUNC_NAME_DATEID = 'func_name.dateid'
    FUNC_NAME_KWARGS = 'func_name.kwargs'
    # FUNC_NAME_KWARGS_DATEID = 'func_name.kwargs.dateid'


class FileCacheMode(Enum):
    DISABLED = 'd'
    OVERWRITE = 'w'
    ENABLED = 'e'
    READONLY = 'r'


class Color(Enum):
    DARK_BLUE = '#002c56'
    MID_BLUE = '#1F4E78'
    LIGHT_BLUE = '#2e75b6'
    ORANGE = '#ff6328'
    GREY = '#939598'
    BLACK = '#000000'
    WHITE = '#FFFFFF'


class EntityType(Enum):
    ACUG = 'AllocatedCapitalUnitGroup'
    ACU = 'AllocatedCapitalUnit'
    STRAT = 'Strategy'


class DBType(Enum):
    MSSQL = 'MSSQL'
    REDSHIFT = 'REDSHIFT'
    POSTGRES = 'POSTGRES'
    MYSQL = 'MYSQL'


class Symbology(Enum):
    AXIOMA_ID = 'AxiomaId'
    BARRA_ID = 'BarraId'
    FIGI = 'FIGI'
    CFIGI = 'CFIGI'
    DISPLAY_CODE = 'DisplayCode'
    TICKER = 'Ticker'
    TICKER_EXCH = 'TickerExch'


class FactorIdentifier(Enum):
    FACTOR_ID = 'FactorId'
    FACTOR_DESC = 'FactorDesc'
    FACTOR_DESC_ID = 'FactorDesc=FactorId'
    FACTOR_CODE = 'FactorCode'


class UniverseMnemonic(Enum):
    SP500 = 'SP500'
    DOW30 = 'DOW30'
    NASDAQ100 = 'NASDAQ100'
    RUSSELL1000 = 'R1000'
    RUSSELL2000 = 'R2000'
    RUSSELL3000 = 'R3000'


class OffsetDateType(Enum):
    CURR_MONTH_END = 'CME'
    NEXT_MONTH_END = 'NME'


class TradingRegion(Enum):
    US = 'US'
    EU = 'EU'
    AP = 'Asia'
    ALL = 'ALL'
    RAW = 'Raw'


class CrossSectionalDataFormat(Enum):
    MATRIX = 'matrix'
    NARROW = 'narrow'


class BasketDecompMode(Enum):
    TOP_DOWN = 'lookthru'
    BOTTOM_UP = 'BOTTOM_UP'



from enum import Enum, unique


# ALPHA STORE
@unique
class ConfigsAttribution(Enum):
    BQ_SRF = 'BQ_SRF'
    BQ_SIM = 'BQ_SIM'


@unique
class ServerType(Enum):
    PROD = 'URL_PROD'
    DEV = 'URL_DEV'
    UAT = 'URL_UAT'
    

@unique
class AssetClass(Enum):
    CREDIT = 'CREDIT'
    EQY = 'EQY'
    FI = 'FI'  #: Fixed Income
    FUT = 'FUT'
    FUTOUTRIGHT = 'FUTOUTRIGHT'
    FX = 'FX'  #: Currencies
    OPTIONS = 'OPTIONS'
    EQY_FUT = 'EQY_FUT'  #: Combined Equity and Futures
    EQY_FUT_FX = 'EQY_FUT_FX'  #: Combined Equity, Futures and FX
    EQY_FX = 'EQY_FX'  #: Combined Equity and FX
    FUT_FX = 'FUT_FX'  #: Combined Futures and FX


@unique
class Frequency(Enum):
    D1 = 'D1'
    D0 = 'D0'
    INT_1min = 'INT_1min'
    INT_5min = 'INT_5min'
    EVENT = 'EVENT'
    IRREGULAR = 'IRREGULAR'
    HOURLY = 'HOURLY'


@unique
class Region(Enum):
    US = 'US'
    EU = 'EU'
    AS = 'AS'
    GL = 'GL'
    EU_ARIES_ID = 'EU_ARIES_ID'


@unique
class RiskModel(Enum):
    AXUS4_MH = 'AXUS4_MH'
    AXUS4_SH = 'AXUS4_SH'
    BARRA_USSLOWS = 'BARRA_USSLOWS'
    BARRA_USSINTM1L = 'BARRA_USSINTM1L'
    BARRA_USFASTD = 'BARRA_USFASTD'
    AXEU4_MH = 'AXEU4_MH'
    AXEU4_SH = 'AXEU4_SH'
    AXEU4_SH_ARIES_ID = 'AXEU4_SH_ARIES_ID'
    AXAP21_MH = 'AXAP21_MH'
    AXAP21_SH = 'AXAP21_SH'
    AXAP4_MH = 'AXAP4_MH'
    AXAP4_SH = 'AXAP4_SH'
    BARRA_ASE2S = 'BARRA_ASE2S'
    AXWW4_MH = 'AXWW4_MH'
    AXWW21_MH = 'AXWW21_MH'
    BARRA_GEMLTS = 'BARRA_GEMLTS'
    NONE = 'NONE'


@unique
class IdsType(Enum):
    CFIGI = 'CFIGI'
    CUSIP = 'CUSIP'
    SEDOL = 'SEDOL'
    TICKER = 'TICKER'
    ARIESID = 'ARIESID'
    SECURITYID = 'SECURITYID'
    MIXED = 'MIXED' 


# NB! Duplicated
@unique
class SignalType(Enum):
    """
    Units/meaning for the VALUE field in an alpha
    """
    SIGNAL = 'SIGNAL'
    POSITION = 'POSITION'
    SHARES = 'SHARES'
    ORDERS = 'ORDERS'
    

@unique
class UploadMethod(Enum):
    """
    Upload methods:
    insert - creates an alpha if it does not exist
    update - updates existing alpha
    overwrite - overwrites the existing alpha or creates one if it does not exist
    """
    insert = 'insert'
    update = 'update'
    overwrite = 'overwrite'


@unique
class DestinationType(Enum):
    staging = 'staging'
    prod = 'prod'
    

@unique 
class Role(Enum):
    User = 'user'
    Superuser = 'superuser'
    AlphaCapture = 'alpha_capture'
    Admin = 'admin'
    

@unique
class Status(Enum):
    RESEARCH = 'RESEARCH'
    TESTING = 'TESTING'
    LIVE = 'LIVE'
    DEPRECATED = 'DEPRECATED'


@unique
class SlippageType(Enum):
    MarketImpact = 'market_impact'
    StaticSlippage = 'static_slippage'


@unique
class ExecutionType(Enum):
    Close = 'Close'
    CloseAuctions = 'CloseAuctions'
    Open = 'Open'
    OpenAuctions = 'OpenAuctions'
    Vwap = 'Vwap'


@unique
class CostModelType(Enum):
    Equity = 'equity'
    Macro = 'macro'


@unique
class GicsLevel(Enum):
    Sector = 1
    Industry = 2
    Subindustry = 3
    
    
@unique
class SignalValueType(Enum):
    """
    Units/meaning for the VALUE field in an alpha.

    For bq is currently implemented for MoneyAllocation only.

    For bq-sim either MoneyAllocation or ExpectedReturns may be used.
    
    Money allocation means the target positions in US dollars.
    """
    MoneyAllocation = 'money_allocation'
    ExpectedReturns = 'expected_returns'
    Ranked = 'ranked'


@unique
class MissingHandlingType(Enum):
    Drift = 'drift'
    Error = 'error'
    Constant = 'constant'
    Zero = 'zero'
    Ignore = 'ignore'


@unique
class IntradayGMVOption(Enum):
    Max = 'max'
    EndOfDay = 'last'
    Mean = 'mean'


@unique
class FactorGroup(Enum):
    Style = 'Style'
    Industry = 'Industry'
    Market = 'Market'
    CustomFactor = 'Custom Factor'
    Country = 'Country'


# BQ_SIM
@unique
class FillType(Enum):
    Close = 'close'
    Vwap = 'vwap'


@unique
class OptimizerType(Enum):
    BAMOptimizer = 'bam_optimizer'
    Mosek = 'MOSEK'
    OSQP = 'OSQP'
    Blend = 'blend'


@unique
class DataLocationEnvironment(Enum):
    """
    Data location environment

    :param Enum: _description_
    :type Enum: _type_
    """
    Prod = 'prod'
    Dev = 'dev'
    Live = 'live'
    Test = 'test'

@unique
class RiskDataTimestamp(Enum):
    pos_t = 0
    pos_t_lag1 = 1
    pos_t_lag2 = 2
